# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
from operator import itemgetter

from pyomo.common.deprecation import deprecation_warning
from pyomo.common.numeric_types import (
    native_types,
    native_numeric_types,
    native_complex_types,
)
from pyomo.core.expr.numeric_expr import (
    NegationExpression,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    AbsExpression,
    UnaryFunctionExpression,
    Expr_ifExpression,
    MonomialTermExpression,
    LinearExpression,
    SumExpression,
    ExternalFunctionExpression,
    mutable_expression,
)
from pyomo.core.expr.relational_expr import (
    EqualityExpression,
    InequalityExpression,
    RangedExpression,
)
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor, _EvaluationVisitor
from pyomo.core.expr import is_fixed, value
from pyomo.core.base.expression import Expression
import pyomo.core.kernel as kernel
from pyomo.repn.util import (
    BeforeChildDispatcher,
    ExitNodeDispatcher,
    ExprType,
    FileDeterminism,
    FileDeterminism_to_SortComponents,
    InvalidNumber,
    OrderedVarRecorder,
    VarRecorder,
    apply_node_operation,
    complex_number_error,
    initialize_exit_node_dispatcher,
    nan,
    sum_like_expression_types,
    val2str,
    check_constant,
)

_CONSTANT = ExprType.CONSTANT
_FIXED = ExprType.FIXED
_VARIABLE = ExprType.VARIABLE
_LINEAR = ExprType.LINEAR
_GENERAL = ExprType.GENERAL


def _merge_dict(dest_dict, src_dict, mult, flag):
    if not src_dict:
        return
    if flag == 1:
        for vid, coef in src_dict.items():
            if vid in dest_dict:
                dest_dict[vid] += coef
            else:
                dest_dict[vid] = coef
    elif not flag:
        # mult is 0.  There is nothing to do, unless the src_dict has an InvalidNumber
        for vid, coef in src_dict.items():
            if coef.__class__ is InvalidNumber:
                if vid in dest_dict:
                    dest_dict[vid] += mult * coef
                else:
                    dest_dict[vid] = mult * coef
    else:
        for vid, coef in src_dict.items():
            if vid in dest_dict:
                dest_dict[vid] += mult * coef
            else:
                dest_dict[vid] = mult * coef


class LinearRepn:
    __slots__ = ("multiplier", "constant", "linear", "nonlinear")

    def __init__(self):
        self.multiplier = 1
        self.constant = 0
        self.linear = {}
        self.nonlinear = None

    def __str__(self):
        linear = (
            "{"
            + ", ".join(f"{val2str(k)}: {val2str(v)}" for k, v in self.linear.items())
            + "}"
        )
        return (
            f"{self.__class__.__name__}(mult={val2str(self.multiplier)}, "
            f"const={val2str(self.constant)}, "
            f"linear={linear}, "
            f"nonlinear={self.nonlinear})"
        )

    def __repr__(self):
        return str(self)

    @staticmethod
    def constant_flag(val):
        return val

    @staticmethod
    def multiplier_flag(val):
        return val

    def walker_exitNode(self):
        if self.nonlinear is not None:
            return _GENERAL, self
        elif self.linear:
            return _LINEAR, self
        else:
            return _CONSTANT, self.multiplier * self.constant

    def duplicate(self):
        ans = self.__class__.__new__(self.__class__)
        ans.multiplier = self.multiplier
        ans.constant = self.constant
        ans.linear = dict(self.linear)
        ans.nonlinear = self.nonlinear
        return ans

    def to_expression(self, visitor):
        if self.nonlinear is not None:
            # We want to start with the nonlinear term (and use
            # assignment to ans instead of addition) in case the term is
            # a non-numeric node (like a relational expression)
            ans = self.nonlinear
        else:
            ans = 0
        with mutable_expression() as e:
            if self.linear:
                var_map = visitor.var_map
                for vid, coef in self.linear.items():
                    if self.multiplier_flag(coef):
                        e += coef * var_map[vid]
            if self.constant_flag(self.constant):
                e += self.constant
        if e.nargs() > 1:
            ans += e
        elif e.nargs() == 1:
            ans += e.arg(0)
        if self.multiplier_flag(self.multiplier) != 1:
            ans *= self.multiplier
        return ans

    def append(self, other):
        """Append a child result from acceptChildResult

        Notes
        -----
        This method assumes that the operator was "+". It is implemented
        so that we can directly use a LinearRepn() as a `data` object in
        the expression walker (thereby allowing us to use the default
        implementation of acceptChildResult [which calls
        `data.append()`] and avoid the function call for a custom
        callback).

        """
        # Note that self.multiplier will always be 1 (we only call append()
        # within a sum, so there is no opportunity for self.multiplier to
        # change). Omitting the assertion for efficiency.
        # assert self.multiplier == 1
        _type, other = other
        if _type <= _FIXED:  # Note: catching _FIXED and _CONSTANT
            self.constant += other
            return

        other_mult_flag = self.multiplier_flag(other.multiplier)
        if other_mult_flag == 1:
            self.constant += other.constant
            if other.linear:
                for vid, coef in other.linear.items():
                    if vid in self.linear:
                        self.linear[vid] += coef
                    else:
                        self.linear[vid] = coef
            if other.nonlinear is not None:
                if self.nonlinear is None:
                    self.nonlinear = other.nonlinear
                else:
                    self.nonlinear += other.nonlinear
            return
        mult = other.multiplier
        if not other_mult_flag:
            # 0 * other, so you would think that there is nothing to
            # add/change about self.  However, there is a chance
            # that other contains an InvalidNumber, so we should go
            # looking for it...
            if other.constant.__class__ is InvalidNumber:
                self.constant += mult * other.constant
            for vid, coef in other.linear.items():
                if coef.__class__ is InvalidNumber:
                    if vid in self.linear:
                        self.linear[vid] += mult * coef
                    else:
                        self.linear[vid] = mult * coef
        else:
            #
            # mult != 0 or 1
            #
            if self.constant_flag(other.constant):
                self.constant += mult * other.constant
            if other.linear:
                for vid, coef in other.linear.items():
                    if vid in self.linear:
                        self.linear[vid] += mult * coef
                    else:
                        self.linear[vid] = mult * coef
        if other.nonlinear is not None:
            if self.nonlinear is None:
                self.nonlinear = mult * other.nonlinear
            else:
                self.nonlinear += mult * other.nonlinear


def to_expression(visitor, arg):
    if arg[0] <= _VARIABLE:  # Note: catching _VARIABLE, _FIXED, and _CONSTANT
        return arg[1]
    else:
        return arg[1].to_expression(visitor)


#
# NEGATION handlers
#


def _handle_negation_constant(visitor, node, arg):
    return (_CONSTANT, -1 * arg[1])


def _handle_negation_fixed(visitor, node, arg):
    return (_FIXED, -1 * arg[1])


def _handle_negation_ANY(visitor, node, arg):
    arg[1].multiplier *= -1
    return arg


#
# PRODUCT handlers
#


def _handle_product_constant_constant(visitor, node, arg1, arg2):
    ans = arg1[1] * arg2[1]
    if ans.__class__ is InvalidNumber:
        constant_flag = visitor.Result.constant_flag
        if not constant_flag(arg1[1]) or not constant_flag(arg2[1]):
            a = val2str(arg1[1])
            b = val2str(arg2[1])
            deprecation_warning(
                f"Encountered {a}*{b} in expression tree.  "
                "Mapping the NaN result to 0 for compatibility "
                "with the lp_v1 writer.  In the future, this NaN "
                "will be preserved/emitted to comply with IEEE-754.",
                version='6.6.0',
            )
            return _CONSTANT, 0
    return _CONSTANT, ans


def _handle_product_fixed_fixed(visitor, node, arg1, arg2):
    # This is valid for fixed * constant, and fixed * fixed
    return _FIXED, arg1[1] * arg2[1]


def _handle_product_constant_ANY(visitor, node, arg1, arg2):
    arg2[1].multiplier *= arg1[1]
    return arg2


def _handle_product_ANY_constant(visitor, node, arg1, arg2):
    arg1[1].multiplier *= arg2[1]
    return arg1


def _handle_product_nonlinear(visitor, node, arg1, arg2):
    # Note: the expectation is that this method is not called often: the
    # Linear visitor is generally expected to be called on linear
    # expressions.  As such, we will not overly concern ourselves with
    # performance here.
    ans = visitor.Result()
    if not visitor.expand_nonlinear_products:
        ans.nonlinear = to_expression(visitor, arg1) * to_expression(visitor, arg2)
        return _GENERAL, ans
    #
    # We are multiplying (and expanding) m(A + Bx + C(x)) * m(A + Bx + C(x))
    _, x1 = arg1
    _, x2 = arg2
    # [mm]
    ans.multiplier = x1.multiplier * x2.multiplier
    # reset the multipliers so that to_expression doesn't re-apply them below
    x1.multiplier = x2.multiplier = 1
    # x1.const * x2.const [AA]
    x1_const_flag = ans.constant_flag(x1.constant)
    x2_const_flag = ans.constant_flag(x2.constant)
    if x1_const_flag and x2_const_flag:
        ans.constant = x1.constant * x2.constant
    # x1.linear * x2.const [BA] + x1.const * x2.linear [AB]
    _merge_dict(ans.linear, x1.linear, x2.constant, x2_const_flag)
    _merge_dict(ans.linear, x2.linear, x1.constant, x1_const_flag)
    NL = 0
    if x2.nonlinear is not None and (
        x1_const_flag or x2.nonlinear.__class__ is InvalidNumber
    ):
        # [AC]
        NL += x1.constant * x2.nonlinear
    if x1.nonlinear is not None and (
        x2_const_flag or x2.nonlinear.__class__ is InvalidNumber
    ):
        # [CA]
        NL += x2.constant * x1.nonlinear
    # [BB] + [BC] + [CB] + [CC]
    x1.constant = 0
    x2.constant = 0
    NL += to_expression(visitor, arg1) * to_expression(visitor, arg2)
    if NL.__class__ in sum_like_expression_types and NL.nargs() == 1:
        NL = NL.arg(0)
    ans.nonlinear = NL
    return _GENERAL, ans


#
# DIVISION handlers
#


def _handle_division_constant_constant(visitor, node, arg1, arg2):
    return _CONSTANT, apply_node_operation(node, (arg1[1], arg2[1]))


def _handle_division_fixed_fixed(visitor, node, arg1, arg2):
    return _FIXED, arg1[1] / arg2[1]


def _handle_division_ANY_constant(visitor, node, arg1, arg2):
    repn = arg1[1]
    # We can only apply the division operation (and reduce the
    # multiplier to a native value) if both the multiplier is a native
    # value AND the divisor is a constant.  We know the latter is true
    # here, but must check the former.  There is also a special case if
    # the divisor is 0: then we can reduce the multiplier to an
    # InvalidNumber (using apply_node_operation) regardless of what the
    # dividend is.  Again, note that arg2 is a constant, so we can check
    # it for 0 with bool()
    if repn.multiplier.__class__ in native_numeric_types or not arg2[1]:
        repn.multiplier = apply_node_operation(node, (repn.multiplier, arg2[1]))
    else:
        repn.multiplier /= arg2[1]
    return arg1


def _handle_division_ANY_fixed(visitor, node, arg1, arg2):
    arg1[1].multiplier /= arg2[1]
    return arg1


def _handle_division_nonlinear(visitor, node, arg1, arg2):
    ans = visitor.Result()
    ans.nonlinear = to_expression(visitor, arg1) / to_expression(visitor, arg2)
    return _GENERAL, ans


#
# EXPONENTIATION handlers
#


def _handle_pow_constant_constant(visitor, node, arg1, arg2):
    ans = apply_node_operation(node, (arg1[1], arg2[1]))
    if ans.__class__ in native_complex_types:
        ans = complex_number_error(ans, visitor, node)
    return _CONSTANT, ans


def _handle_pow_fixed_fixed(visitor, node, arg1, arg2):
    return _FIXED, arg1[1] ** arg2[1]


def _handle_pow_ANY_constant(visitor, node, arg1, arg2):
    _, exp = arg2
    if exp == 1:
        return arg1
    elif exp > 1 and exp <= visitor.max_exponential_expansion and int(exp) == exp:
        _type, _arg = arg1
        ans = _type, _arg.duplicate()
        for i in range(1, int(exp)):
            ans = visitor.exit_node_dispatcher[(ProductExpression, ans[0], _type)](
                visitor, None, ans, (_type, _arg.duplicate())
            )
        return ans
    elif not exp:
        return _CONSTANT, 1
    return _handle_pow_nonlinear(visitor, node, arg1, arg2)


def _handle_pow_nonlinear(visitor, node, arg1, arg2):
    ans = visitor.Result()
    ans.nonlinear = to_expression(visitor, arg1) ** to_expression(visitor, arg2)
    return _GENERAL, ans


#
# ABS and UNARY handlers
#


def _handle_unary_constant(visitor, node, arg):
    ans = apply_node_operation(node, (arg[1],))
    # Unary includes sqrt() which can return complex numbers
    if ans.__class__ in native_complex_types:
        ans = complex_number_error(ans, visitor, node)
    return _CONSTANT, ans


def _handle_unary_fixed(visitor, node, arg):
    return _FIXED, node.create_node_with_local_data((arg[1],))


def _handle_unary_nonlinear(visitor, node, arg):
    ans = visitor.Result()
    ans.nonlinear = node.create_node_with_local_data((to_expression(visitor, arg),))
    return _GENERAL, ans


#
# NAMED EXPRESSION handlers
#


def _handle_named_constant(visitor, node, arg1):
    # Record this common expression
    visitor.subexpression_cache[id(node)] = arg1
    return arg1


def _handle_named_ANY(visitor, node, arg1):
    # Record this common expression
    visitor.subexpression_cache[id(node)] = arg1
    _type, arg1 = arg1
    return _type, arg1.duplicate()


#
# EXPR_IF handlers
#


def _handle_expr_if_const(visitor, node, arg1, arg2, arg3):
    _type, _test = arg1
    assert _type is _CONSTANT
    if _test:
        if _test.__class__ is InvalidNumber:
            # nan
            return _handle_expr_if_nonlinear(visitor, node, arg1, arg2, arg3)
        return arg2
    else:
        return arg3


def _handle_expr_if_nonlinear(visitor, node, arg1, arg2, arg3):
    # Note: guaranteed that arg1 is not _CONSTANT
    ans = visitor.Result()
    ans.nonlinear = Expr_ifExpression(
        (
            to_expression(visitor, arg1),
            to_expression(visitor, arg2),
            to_expression(visitor, arg3),
        )
    )
    return _GENERAL, ans


#
# Relational expression handlers
#


def _handle_equality_const(visitor, node, arg1, arg2):
    # It is exceptionally likely that if we get here, one of the
    # arguments is an InvalidNumber
    args, causes = InvalidNumber.parse_args(arg1[1], arg2[1])
    try:
        ans = args[0] == args[1]
    except:
        ans = False
        causes.append(str(sys.exc_info()[1]))
    if causes:
        ans = InvalidNumber(ans, causes)
    return _CONSTANT, ans


def _handle_equality_general(visitor, node, arg1, arg2):
    ans = visitor.Result()
    ans.nonlinear = EqualityExpression(
        (to_expression(visitor, arg1), to_expression(visitor, arg2))
    )
    return _GENERAL, ans


def _handle_inequality_const(visitor, node, arg1, arg2):
    # It is exceptionally likely that if we get here, one of the
    # arguments is an InvalidNumber
    args, causes = InvalidNumber.parse_args(arg1[1], arg2[1])
    try:
        ans = args[0] <= args[1]
    except:
        ans = False
        causes.append(str(sys.exc_info()[1]))
    if causes:
        ans = InvalidNumber(ans, causes)
    return _CONSTANT, ans


def _handle_inequality_general(visitor, node, arg1, arg2):
    ans = visitor.Result()
    ans.nonlinear = InequalityExpression(
        (to_expression(visitor, arg1), to_expression(visitor, arg2)), node.strict
    )
    return _GENERAL, ans


def _handle_ranged_const(visitor, node, arg1, arg2, arg3):
    # It is exceptionally likely that if we get here, one of the
    # arguments is an InvalidNumber
    args, causes = InvalidNumber.parse_args(arg1[1], arg2[1], arg3[1])
    try:
        ans = args[0] <= args[1] <= args[2]
    except:
        ans = False
        causes.append(str(sys.exc_info()[1]))
    if causes:
        ans = InvalidNumber(ans, causes)
    return _CONSTANT, ans


def _handle_ranged_general(visitor, node, arg1, arg2, arg3):
    ans = visitor.Result()
    ans.nonlinear = RangedExpression(
        (
            to_expression(visitor, arg1),
            to_expression(visitor, arg2),
            to_expression(visitor, arg3),
        ),
        node.strict,
    )
    return _GENERAL, ans


def define_exit_node_handlers(_exit_node_handlers=None):
    if _exit_node_handlers is None:
        _exit_node_handlers = {}
    _exit_node_handlers[NegationExpression] = {
        None: _handle_negation_ANY,
        (_CONSTANT,): _handle_negation_constant,
        (_FIXED,): _handle_negation_fixed,
    }
    _exit_node_handlers[ProductExpression] = {
        None: _handle_product_nonlinear,
        (_CONSTANT, _CONSTANT): _handle_product_constant_constant,
        (_CONSTANT, _FIXED): _handle_product_fixed_fixed,
        (_CONSTANT, _LINEAR): _handle_product_constant_ANY,
        (_CONSTANT, _GENERAL): _handle_product_constant_ANY,
        (_FIXED, _CONSTANT): _handle_product_fixed_fixed,
        (_FIXED, _FIXED): _handle_product_fixed_fixed,
        (_FIXED, _LINEAR): _handle_product_constant_ANY,
        (_FIXED, _GENERAL): _handle_product_constant_ANY,
        (_LINEAR, _CONSTANT): _handle_product_ANY_constant,
        (_LINEAR, _FIXED): _handle_product_ANY_constant,
        (_GENERAL, _CONSTANT): _handle_product_ANY_constant,
        (_GENERAL, _FIXED): _handle_product_ANY_constant,
    }
    _exit_node_handlers[MonomialTermExpression] = _exit_node_handlers[ProductExpression]
    _exit_node_handlers[DivisionExpression] = {
        None: _handle_division_nonlinear,
        (_CONSTANT, _CONSTANT): _handle_division_constant_constant,
        (_CONSTANT, _FIXED): _handle_division_fixed_fixed,
        (_FIXED, _CONSTANT): _handle_division_fixed_fixed,
        (_FIXED, _FIXED): _handle_division_fixed_fixed,
        (_LINEAR, _CONSTANT): _handle_division_ANY_constant,
        (_LINEAR, _FIXED): _handle_division_ANY_fixed,
        (_GENERAL, _CONSTANT): _handle_division_ANY_constant,
        (_GENERAL, _FIXED): _handle_division_ANY_fixed,
    }
    _exit_node_handlers[PowExpression] = {
        None: _handle_pow_nonlinear,
        (_CONSTANT, _CONSTANT): _handle_pow_constant_constant,
        (_CONSTANT, _FIXED): _handle_pow_fixed_fixed,
        (_FIXED, _CONSTANT): _handle_pow_fixed_fixed,
        (_FIXED, _FIXED): _handle_pow_fixed_fixed,
        (_LINEAR, _CONSTANT): _handle_pow_ANY_constant,
        (_GENERAL, _CONSTANT): _handle_pow_ANY_constant,
    }
    _exit_node_handlers[UnaryFunctionExpression] = {
        None: _handle_unary_nonlinear,
        (_CONSTANT,): _handle_unary_constant,
        (_FIXED,): _handle_unary_fixed,
    }
    _exit_node_handlers[AbsExpression] = _exit_node_handlers[UnaryFunctionExpression]
    _exit_node_handlers[Expression] = {
        None: _handle_named_ANY,
        (_CONSTANT,): _handle_named_constant,
        (_FIXED,): _handle_named_constant,
    }
    _exit_node_handlers[Expr_ifExpression] = {None: _handle_expr_if_nonlinear}
    for j in (_CONSTANT, _LINEAR, _GENERAL):
        for k in (_CONSTANT, _LINEAR, _GENERAL):
            _exit_node_handlers[Expr_ifExpression][
                _CONSTANT, j, k
            ] = _handle_expr_if_const
    _exit_node_handlers[EqualityExpression] = {
        None: _handle_equality_general,
        (_CONSTANT, _CONSTANT): _handle_equality_const,
    }
    _exit_node_handlers[InequalityExpression] = {
        None: _handle_inequality_general,
        (_CONSTANT, _CONSTANT): _handle_inequality_const,
    }
    _exit_node_handlers[RangedExpression] = {
        None: _handle_ranged_general,
        (_CONSTANT, _CONSTANT, _CONSTANT): _handle_ranged_const,
    }
    return _exit_node_handlers


class LinearBeforeChildDispatcher(BeforeChildDispatcher):
    def __init__(self):
        # Special handling for external functions: will be handled
        # as terminal nodes from the point of view of the visitor
        self[ExternalFunctionExpression] = self._before_external
        # Special linear / summation expressions
        self[MonomialTermExpression] = self._before_monomial
        self[LinearExpression] = self._before_linear
        self[SumExpression] = self._before_general_expression

    @staticmethod
    def _before_var(visitor, child):
        _id = id(child)
        if _id not in visitor.var_map:
            if child.fixed:
                return False, (_CONSTANT, check_constant(child.value, child, visitor))
            visitor.var_recorder.add(child)
        ans = visitor.Result()
        ans.linear[_id] = 1
        return False, (_LINEAR, ans)

    @staticmethod
    def _before_monomial(visitor, child):
        #
        # The following are performance optimizations for common
        # situations (Monomial terms and Linear expressions)
        #
        arg1, arg2 = child._args_
        if arg1.__class__ not in native_types:
            try:
                arg1 = check_constant(visitor.evaluate(arg1), arg1, visitor)
            except (ValueError, ArithmeticError):
                return True, None

        # We want to check / update the var_map before processing "0"
        # coefficients so that we are consistent with what gets added to the
        # var_map (e.g., 0*x*y: y is processed by _before_var and will
        # always be added, but x is processed here)
        _id = id(arg2)
        if _id not in visitor.var_map:
            if arg2.fixed:
                return False, (
                    _CONSTANT,
                    arg1 * check_constant(arg2.value, arg2, visitor),
                )
            visitor.var_recorder.add(arg2)

        # Trap multiplication by 0 and nan.  Note that arg1 was reduced
        # to a numeric value at the beginning of this method.
        if not arg1:
            if arg2.fixed:
                arg2 = check_constant(arg2.value, arg2, visitor)
                if arg2.__class__ is InvalidNumber:
                    deprecation_warning(
                        f"Encountered {arg1}*{val2str(arg2)} in expression "
                        "tree.  Mapping the NaN result to 0 for compatibility "
                        "with the lp_v1 writer.  In the future, this NaN "
                        "will be preserved/emitted to comply with IEEE-754.",
                        version='6.6.0',
                    )
            return False, (_CONSTANT, arg1)

        ans = visitor.Result()
        ans.linear[_id] = arg1
        return False, (_LINEAR, ans)

    @staticmethod
    def _before_linear(visitor, child):
        var_map = visitor.var_map
        ans = visitor.Result()
        const = 0
        linear = ans.linear
        for arg in child.args:
            if arg.__class__ is MonomialTermExpression:
                arg1, arg2 = arg._args_
                if arg1.__class__ not in native_types:
                    try:
                        arg1 = check_constant(visitor.evaluate(arg1), arg1, visitor)
                    except (ValueError, ArithmeticError):
                        return True, None

                # Trap multiplication by 0 and nan.  Note that arg1 was
                # reduced to a numeric value at the beginning of this
                # method.
                if not arg1:
                    if arg2.fixed:
                        arg2 = check_constant(arg2.value, arg2, visitor)
                        if arg2.__class__ is InvalidNumber:
                            deprecation_warning(
                                f"Encountered {arg1}*{val2str(arg2)} in expression "
                                "tree.  Mapping the NaN result to 0 for compatibility "
                                "with the lp_v1 writer.  In the future, this NaN "
                                "will be preserved/emitted to comply with IEEE-754.",
                                version='6.6.0',
                            )
                    continue

                _id = id(arg2)
                if _id not in var_map:
                    if arg2.fixed:
                        const += arg1 * check_constant(arg2.value, arg2, visitor)
                        continue
                    visitor.var_recorder.add(arg2)
                    linear[_id] = arg1
                elif _id in linear:
                    linear[_id] += arg1
                else:
                    linear[_id] = arg1
            elif arg.__class__ in native_numeric_types:
                const += arg
            elif arg.is_variable_type():
                _id = id(arg)
                if _id not in var_map:
                    if arg.fixed:
                        const += check_constant(arg.value, arg, visitor)
                        continue
                    visitor.var_recorder.add(arg)
                    linear[_id] = 1
                elif _id in linear:
                    linear[_id] += 1
                else:
                    linear[_id] = 1
            else:
                try:
                    const += check_constant(visitor.evaluate(arg), arg, visitor)
                except (ValueError, ArithmeticError):
                    return True, None
        if linear:
            ans.constant = const
            return False, (_LINEAR, ans)
        else:
            return False, (_CONSTANT, const)

    @staticmethod
    def _before_named_expression(visitor, child):
        _id = id(child)
        if _id in visitor.subexpression_cache:
            _type, expr = visitor.subexpression_cache[_id]
            if _type is _CONSTANT:
                return False, (_type, expr)
            else:
                return False, (_type, expr.duplicate())
        else:
            return True, None

    @staticmethod
    def _before_external(visitor, child):
        ans = visitor.Result()
        if all(is_fixed(arg) for arg in child.args):
            try:
                ans.constant = check_constant(visitor.evaluate(child), child, visitor)
                return False, (_CONSTANT, ans)
            except:
                pass
        ans.nonlinear = child
        return False, (_GENERAL, ans)


class LinearRepnVisitor(StreamBasedExpressionVisitor):
    Result = LinearRepn
    before_child_dispatcher = LinearBeforeChildDispatcher()
    exit_node_dispatcher = ExitNodeDispatcher(
        initialize_exit_node_dispatcher(define_exit_node_handlers())
    )
    expand_nonlinear_products = False
    max_exponential_expansion = 1

    def __init__(
        self,
        subexpression_cache,
        var_map=None,
        var_order=None,
        sorter=None,
        var_recorder=None,
    ):
        super().__init__()
        self.subexpression_cache = subexpression_cache
        if any(_ is not None for _ in (var_map, var_order, sorter)):
            if var_recorder is not None:
                raise ValueError(
                    "LinearRepnVisitor: cannot specify any of var_map, "
                    "var_order, or sorter with var_recorder"
                )
            deprecation_warning(
                "var_map, var_order, and sorter are deprecated arguments to "
                "LinearRepnVisitor().  Please pass the VarRecorder object directly.",
                version='6.8.1',
            )
            var_recorder = OrderedVarRecorder(var_map, var_order, sorter)
        if var_recorder is None:
            var_recorder = VarRecorder(
                {}, FileDeterminism_to_SortComponents(FileDeterminism.ORDERED)
            )
        self.var_recorder = var_recorder
        self.var_map = var_recorder.var_map
        self._eval_expr_visitor = _EvaluationVisitor(True)
        self.evaluate = self._eval_expr_visitor.dfs_postorder_stack

    def initializeWalker(self, expr):
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            return False, self.finalizeResult(result)
        return True, expr

    def beforeChild(self, node, child, child_idx):
        return self.before_child_dispatcher[child.__class__](self, child)

    def enterNode(self, node):
        # SumExpression are potentially large nary operators.  Directly
        # populate the result
        if node.__class__ in sum_like_expression_types:
            return node.args, self.Result()
        else:
            return node.args, []

    def exitNode(self, node, data):
        if data.__class__ is self.Result:
            return data.walker_exitNode()
        #
        # General expressions...
        #
        return self.exit_node_dispatcher[(node.__class__, *map(itemgetter(0), data))](
            self, node, *data
        )

    def finalizeResult(self, result):
        ans = result[1]
        if ans.__class__ is not self.Result:
            ans = self.Result()
            assert result[0] <= _FIXED  # Note: allowing _FIXED or _CONSTANT
            ans.constant = result[1]
            return ans

        mult_flag = ans.multiplier_flag(ans.multiplier)
        if mult_flag == 1:
            # mult is identity: only thing to do is filter out zero coefficients
            self._filter_zeros(ans)
            return ans
        elif not mult_flag:
            # the multiplier has cleared out the entire expression.
            # Warn if this is suppressing a NaN (unusual, and
            # non-standard, but we will wait to remove this behavior
            # for the time being)
            if ans.constant.__class__ is InvalidNumber or any(
                c.__class__ is InvalidNumber for c in ans.linear.values()
            ):
                deprecation_warning(
                    f"Encountered {ans.multiplier}*nan in expression tree.  "
                    "Mapping the NaN result to 0 for compatibility "
                    "with the lp_v1 writer.  In the future, this NaN "
                    "will be preserved/emitted to comply with IEEE-754.",
                    version='6.6.0',
                )
            return self.Result()

        # mult not in {0, 1}: factor it into the constant,
        # linear coefficients, and nonlinear term
        self._factor_multiplier_into_ans(ans, ans.multiplier)
        return ans

    def _filter_zeros(self, ans):
        _flag = ans.constant_flag
        # Note: creating the intermediate list is important, as we are
        # modifying the dict in place.
        for vid in [vid for vid, c in ans.linear.items() if not _flag(c)]:
            del ans.linear[vid]

    def _factor_multiplier_into_ans(self, ans, mult):
        _flag = ans.constant_flag
        linear = ans.linear
        zeros = []
        for vid, coef in linear.items():
            prod = mult * coef
            if _flag(prod):
                linear[vid] = prod
            else:
                zeros.append(vid)
        for vid in zeros:
            del linear[vid]
        if ans.nonlinear is not None:
            ans.nonlinear *= mult
        if _flag(ans.constant):
            ans.constant *= mult
        ans.multiplier = 1



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.contrib.solver.solvers.highs import Highs

opt = Highs()
if not opt.available():
    raise unittest.SkipTest


@unittest.pytest.mark.solver("highs")
class TestBugs(unittest.TestCase):
    def test_mutable_params_with_remove_cons(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-10, 10))
        m.y = pyo.Var()

        m.p1 = pyo.Param(mutable=True)
        m.p2 = pyo.Param(mutable=True)

        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= m.x + m.p1)
        m.c2 = pyo.Constraint(expr=m.y >= -m.x + m.p2)

        m.p1.value = 1
        m.p2.value = 1

        opt = Highs()
        res = opt.solve(m)
        self.assertAlmostEqual(res.objective_bound, 1)

        del m.c1
        m.p2.value = 2
        res = opt.solve(m)
        self.assertAlmostEqual(res.objective_bound, -8)

    def test_mutable_params_with_remove_vars(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()

        m.p1 = pyo.Param(mutable=True)
        m.p2 = pyo.Param(mutable=True)

        m.y.setlb(m.p1)
        m.y.setub(m.p2)

        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= m.x + 1)
        m.c2 = pyo.Constraint(expr=m.y >= -m.x + 1)

        m.p1.value = -10
        m.p2.value = 10

        opt = Highs()
        res = opt.solve(m)
        self.assertAlmostEqual(res.objective_bound, 1)

        del m.c1
        del m.c2
        m.p1.value = -9
        m.p2.value = 9
        res = opt.solve(m)
        self.assertAlmostEqual(res.objective_bound, -9)

    def test_fix_and_unfix(self):
        # Tests issue https://github.com/Pyomo/pyomo/issues/3127

        m = pyo.ConcreteModel()
        m.x = pyo.Var(domain=pyo.Binary)
        m.y = pyo.Var(domain=pyo.Binary)
        m.fx = pyo.Var(domain=pyo.NonNegativeReals)
        m.fy = pyo.Var(domain=pyo.NonNegativeReals)
        m.c1 = pyo.Constraint(expr=m.fx <= m.x)
        m.c2 = pyo.Constraint(expr=m.fy <= m.y)
        m.c3 = pyo.Constraint(expr=m.x + m.y <= 1)

        m.obj = pyo.Objective(expr=m.fx * 0.5 + m.fy * 0.4, sense=pyo.maximize)

        opt = Highs()

        # solution 1 has m.x == 1 and m.y == 0
        r = opt.solve(m)
        self.assertAlmostEqual(m.fx.value, 1, places=5)
        self.assertAlmostEqual(m.fy.value, 0, places=5)
        self.assertAlmostEqual(r.objective_bound, 0.5, places=5)

        # solution 2 has m.x == 0 and m.y == 1
        m.y.fix(1)
        r = opt.solve(m)
        self.assertAlmostEqual(m.fx.value, 0, places=5)
        self.assertAlmostEqual(m.fy.value, 1, places=5)
        self.assertAlmostEqual(r.objective_bound, 0.4, places=5)

        # solution 3 should be equal solution 1
        m.y.unfix()
        m.x.fix(1)
        r = opt.solve(m)
        self.assertAlmostEqual(m.fx.value, 1, places=5)
        self.assertAlmostEqual(m.fy.value, 0, places=5)
        self.assertAlmostEqual(r.objective_bound, 0.5, places=5)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
import re
import pyomo.common.unittest as unittest
from pyomo.common.tee import capture_output
from pyomo.common.log import LoggingIntercept

from pyomo.environ import SolverFactory
from pyomo.scripting.driver_help import help_solvers, help_transformations
from pyomo.scripting.pyomo_main import main


def _mock_kestrel(solvers, err=None):
    """
    Context-manager that replaces pyomo.neos.kestrel.kestrelAMPL with a
    stub whose ``solvers()`` method returns `solvers` and whose
    ``connect_error`` attribute is `err`.
    """
    import pyomo.neos.kestrel as _k

    class _MockedNEOS:
        def __init__(self, *a, **kw):
            self._solvers = solvers
            self.connect_error = err

        def solvers(self):
            return self._solvers

    return unittest.mock.patch.object(_k, "kestrelAMPL", _MockedNEOS)


class Test(unittest.TestCase):
    def test_pyomo_main_deprecation(self):
        with LoggingIntercept() as LOG:
            with unittest.pytest.raises(SystemExit) as e:
                main(args=['--solvers=glpk', 'foo.py'])
        self.assertIn("Running the 'pyomo' script with no subcommand", LOG.getvalue())

    def test_help_solvers(self):
        with capture_output() as OUT:
            help_solvers()
        OUT = OUT.getvalue()
        self.assertTrue(re.search('Pyomo Solvers and Solver Managers', OUT))
        self.assertTrue(re.search('Serial Solver', OUT))
        # Test known solvers and metasolver flags
        # ASL is a metasolver
        self.assertTrue(re.search(r'\n   \*asl ', OUT))
        # MindtPY is bundled with Pyomo so should always be available
        self.assertTrue(re.search(r'\n   \+mindtpy ', OUT))
        for solver in ('ipopt', 'cbc', 'glpk'):
            s = SolverFactory(solver)
            if s.available():
                self.assertTrue(
                    re.search(r"\n   \+%s " % solver, OUT),
                    "'   +%s' not found in help --solvers" % solver,
                )
            else:
                self.assertTrue(
                    re.search(r"\n    %s " % solver, OUT),
                    "'    %s' not found in help --solvers" % solver,
                )
        for solver in ('baron',):
            s = SolverFactory(solver)
            if s.license_is_valid():
                self.assertTrue(
                    re.search(r"\n   \+%s " % solver, OUT),
                    "'   +%s' not found in help --solvers" % solver,
                )
            elif s.available():
                self.assertTrue(
                    re.search(r"\n   \-%s " % solver, OUT),
                    "'   -%s' not found in help --solvers" % solver,
                )
            else:
                self.assertTrue(
                    re.search(r"\n    %s " % solver, OUT),
                    "'    %s' not found in help --solvers" % solver,
                )

    def test_help_solvers_neos_available(self):
        with _mock_kestrel(['ipoptAMPL', 'knitroAMPL']), capture_output() as OUT:
            help_solvers()
        txt = OUT.getvalue()
        self.assertIn("NEOS Solver Interfaces", txt)
        self.assertIn("ipopt", txt.lower())
        self.assertIn("knitro", txt.lower())
        self.assertIn("solver interfaces are available", txt)

    def test_help_solvers_neos_unavailable(self):
        """NEOS failure should show a message as to why"""
        import socket, xmlrpc.client

        cases = [
            (NotImplementedError("no ssl"), "SSL support"),
            (socket.timeout(), "timed out"),
            (socket.gaierror(-2, "not known"), "resolved"),
            (
                xmlrpc.client.ProtocolError(
                    "https://neos-server.org", 503, "Service Unavailable", {}
                ),
                "HTTP 503",
            ),
        ]

        for err, output in cases:
            with self.subTest(err=err), _mock_kestrel([], err), capture_output() as OUT:
                help_solvers()
            txt = OUT.getvalue()
            self.assertIn("NEOS Solver Interfaces", txt)
            self.assertIn("currently unavailable", txt)
            self.assertIn(output, txt)

    def test_help_transformations(self):
        with capture_output() as OUT:
            help_transformations()
        OUT = OUT.getvalue()
        self.assertTrue(re.search('Pyomo Model Transformations', OUT))
        self.assertTrue(re.search('core.relax_integer_vars', OUT))
        # test a transformation that we know is deprecated
        self.assertTrue(re.search(r'duality.linear_dual\s+\[DEPRECATED\]', OUT))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    RangeSet,
    ConstraintList,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class LP_duals_minimize(_BaseTestModel):
    """
    A continuous linear model designed to test every form of
    constraint when collecting duals for a minimization
    objective
    """

    description = "LP_duals_minimize"
    level = ('nightly', 'expensive')
    capabilities = set(['linear'])
    size = (12, 12, None)

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = None
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.s = RangeSet(1, 12)
        model.x = Var(model.s)
        model.x[1].setlb(-1)
        model.x[1].setub(1)
        model.x[2].setlb(-1)
        model.x[2].setub(1)
        model.obj = Objective(
            expr=sum(model.x[i] * ((-1) ** (i + 1)) for i in model.x.index_set())
        )
        model.c = ConstraintList()
        # to make the variable used in the constraint match the name
        model.c.add(Constraint.Skip)
        model.c.add(Constraint.Skip)
        model.c.add(model.x[3] >= -1.0)
        model.c.add(model.x[4] <= 1.0)
        model.c.add(model.x[5] == -1.0)
        model.c.add(model.x[6] == -1.0)
        model.c.add(model.x[7] == 1.0)
        model.c.add(model.x[8] == 1.0)
        model.c.add((-1.0, model.x[9], -1.0))
        model.c.add((-1.0, model.x[10], -1.0))
        model.c.add((1.0, model.x[11], 1.0))
        model.c.add((1.0, model.x[12], 1.0))

        model.c_inactive = ConstraintList()
        # to make the variable used in the constraint match the name
        model.c_inactive.add(Constraint.Skip)
        model.c_inactive.add(Constraint.Skip)
        model.c_inactive.add(model.x[3] >= -2.0)
        model.c_inactive.add(model.x[4] <= 2.0)

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        for i in model.s:
            model.x[i].value = None


@register_model
class LP_duals_minimize_kernel(LP_duals_minimize):
    def _generate_model(self):
        self.model = None
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.s = list(range(1, 13))
        model.x = pmo.variable_dict((i, pmo.variable()) for i in model.s)
        model.x[1].lb = -1
        model.x[1].ub = 1
        model.x[2].lb = -1
        model.x[2].ub = 1
        model.obj = pmo.objective(sum(model.x[i] * ((-1) ** (i + 1)) for i in model.s))
        model.c = pmo.constraint_dict()
        # to make the variable used in the constraint match the name
        model.c[3] = pmo.constraint(model.x[3] >= -1.0)
        model.c[4] = pmo.constraint(model.x[4] <= 1.0)
        model.c[5] = pmo.constraint(model.x[5] == -1.0)
        model.c[6] = pmo.constraint(model.x[6] == -1.0)
        model.c[7] = pmo.constraint(model.x[7] == 1.0)
        model.c[8] = pmo.constraint(model.x[8] == 1.0)
        model.c[9] = pmo.constraint((-1.0, model.x[9], -1.0))
        model.c[10] = pmo.constraint((-1.0, model.x[10], -1.0))
        model.c[11] = pmo.constraint((1.0, model.x[11], 1.0))
        model.c[12] = pmo.constraint((1.0, model.x[12], 1.0))

        model.c_inactive = pmo.constraint_dict()
        # to make the variable used in the constraint match the name
        model.c_inactive[3] = pmo.constraint(model.x[3] >= -2.0)
        model.c_inactive[4] = pmo.constraint(model.x[4] <= 2.0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Re-implementation of example 1 of Outer approximation and ECP.

Re-implementation of Duran example 1 as written by Westerlund
MINLP test problem in Pyomo
Author: David Bernal <https://github.com/bernalde>

The expected optimal solution value is 6.00976.

Ref:
    Duran, Marco A., and Ignacio E. Grossmann.
    'An outer-approximation algorithm for a class of mixed-integer nonlinear
    programs.'
    Mathematical programming 36.3 (1986): 307-339.
    Westerlund, Tapio, and Frank Pettersson.
    'An extended cutting plane method for solving convex MINLP problems.'
    Computers & Chemical Engineering 19 (1995): 131-136.
    Example 1

    Problem type:    convex MINLP
            size:    3  binary variables
                     4  continuous variables
                     11  constraints


"""

from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    NonNegativeReals,
    Objective,
    RangeSet,
    Var,
    minimize,
    log,
)
from pyomo.common.collections import ComponentMap


class SimpleMINLP(ConcreteModel):
    """Example 1 Outer Approximation and Extended Cutting Planes."""

    def __init__(self, *args, **kwargs):
        """Create the problem."""
        kwargs.setdefault('name', 'SimpleMINLP2')
        super(SimpleMINLP, self).__init__(*args, **kwargs)
        m = self

        """Set declarations"""
        I = m.I = RangeSet(1, 4, doc='continuous variables')
        J = m.J = RangeSet(1, 3, doc='discrete variables')

        # initial point information for discrete variables
        initY = {1: 1, 2: 0, 3: 1}
        # initial point information for continuous variables
        initX = {1: 0, 2: 0, 3: 0, 4: 0}

        """Variable declarations"""
        # DISCRETE VARIABLES
        Y = m.Y = Var(J, domain=Binary, initialize=initY)
        # CONTINUOUS VARIABLES
        X = m.X = Var(I, domain=NonNegativeReals, initialize=initX, bounds=(0, 2))

        """Constraint definitions"""
        # CONSTRAINTS
        m.const1 = Constraint(
            expr=0.8 * log(X[2] + 1) + 0.96 * log(X[1] - X[2] + 1) - 0.8 * X[3] >= 0
        )
        m.const2 = Constraint(
            expr=log(X[2] + 1) + 1.2 * log(X[1] - X[2] + 1) - X[3] - 2 * Y[3] >= -2
        )
        m.const3 = Constraint(
            expr=10 * X[1]
            - 7 * X[3]
            - 18 * log(X[2] + 1)
            - 19.2 * log(X[1] - X[2] + 1)
            + 10
            - X[4]
            <= 0
        )
        m.const4 = Constraint(expr=X[2] - X[1] <= 0)
        m.const5 = Constraint(expr=X[2] - 2 * Y[1] <= 0)
        m.const6 = Constraint(expr=X[1] - X[2] - 2 * Y[2] <= 0)
        m.const7 = Constraint(expr=Y[1] + Y[2] <= 1)

        """Cost (objective) function definition"""
        m.objective = Objective(
            expr=+5 * Y[1] + 6 * Y[2] + 8 * Y[3] + X[4], sense=minimize
        )

        """Bound definitions"""
        # x (continuous) upper bounds
        x_ubs = {1: 2, 2: 2, 3: 1, 4: 100}
        for i, x_ub in x_ubs.items():
            X[i].setub(x_ub)

        m.optimal_value = 6.00976
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.X[1]] = 1.3009758908698426
        m.optimal_solution[m.X[2]] = 0.0
        m.optimal_solution[m.X[3]] = 1.0
        m.optimal_solution[m.X[4]] = 0.009758908698423729
        m.optimal_solution[m.Y[1]] = 0.0
        m.optimal_solution[m.Y[2]] = 1.0
        m.optimal_solution[m.Y[3]] = 0.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import datetime
import multiprocessing
import os
import pickle
import time

import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept
from pyomo.common.tee import capture_output
from pyomo.common.tempfiles import TempfileManager
from pyomo.environ import ConcreteModel, Var, Param


@unittest.timeout(10)
def short_sleep():
    return 42


@unittest.timeout(0.01)
def long_sleep():
    time.sleep(1)
    return 42


@unittest.timeout(10)
def raise_exception():
    foo.bar


@unittest.timeout(10)
def fail():
    raise AssertionError("0 != 1")


class TestPyomoUnittest(unittest.TestCase):
    def test_assertStructuredAlmostEqual_comparison(self):
        a = 1
        b = 1
        self.assertStructuredAlmostEqual(a, b)
        # default relative tolerance is 1e-7.  This should have a reltol
        # of "exactly" 1e-7, but due to roundoff error, it is not.  The
        # choice of 9.999e-8 and 9.999-e7 is specifically so that
        # roundoff error doesn't cause tests to fail
        b -= 9.999e-8
        self.assertStructuredAlmostEqual(a, b)
        b -= 9.999e-8
        with self.assertRaisesRegex(self.failureException, '1 !~= 0.9999'):
            self.assertStructuredAlmostEqual(a, b)

        b = 1
        self.assertStructuredAlmostEqual(a, b, reltol=1e-6)
        b -= 9.999e-7
        self.assertStructuredAlmostEqual(a, b, reltol=1e-6)
        b -= 9.999e-7
        with self.assertRaisesRegex(self.failureException, '1 !~= 0.999'):
            self.assertStructuredAlmostEqual(a, b, reltol=1e-6)

        b = 1
        self.assertStructuredAlmostEqual(a, b, places=6)
        b -= 9.999e-7
        self.assertStructuredAlmostEqual(a, b, places=6)
        b -= 9.999e-7
        with self.assertRaisesRegex(self.failureException, '1 !~= 0.999'):
            self.assertStructuredAlmostEqual(a, b, places=6)

        with self.assertRaisesRegex(self.failureException, '10 !~= 10.01'):
            self.assertStructuredAlmostEqual(10, 10.01, abstol=1e-3)
        self.assertStructuredAlmostEqual(10, 10.01, reltol=1e-3)
        with self.assertRaisesRegex(self.failureException, '10 !~= 10.01'):
            self.assertStructuredAlmostEqual(10, 10.01, delta=1e-3)

    def test_assertStructuredAlmostEqual_nan(self):
        a = float('nan')
        b = float('nan')
        self.assertStructuredAlmostEqual(a, b)

    def test_assertStructuredAlmostEqual_errorChecking(self):
        with self.assertRaisesRegex(
            ValueError, "Cannot specify more than one of {places, delta, abstol}"
        ):
            self.assertStructuredAlmostEqual(1, 1, places=7, delta=1)

    def test_assertStructuredAlmostEqual_str(self):
        self.assertStructuredAlmostEqual("hi", "hi")
        with self.assertRaisesRegex(self.failureException, "'hi' !~= 'hello'"):
            self.assertStructuredAlmostEqual("hi", "hello")
        with self.assertRaisesRegex(self.failureException, r"'hi' !~= \['h',"):
            self.assertStructuredAlmostEqual("hi", ['h', 'i'])

    def test_assertStructuredAlmostEqual_othertype(self):
        a = datetime.datetime(1, 1, 1)
        b = datetime.datetime(1, 1, 1)
        self.assertStructuredAlmostEqual(a, b)
        b = datetime.datetime(1, 1, 2)
        with self.assertRaisesRegex(self.failureException, "datetime.* !~= datetime"):
            self.assertStructuredAlmostEqual(a, b)

    def test_assertStructuredAlmostEqual_list(self):
        a = [1, 2]
        b = [1, 2, 3]
        with self.assertRaisesRegex(
            self.failureException, r'sequences are different sizes \(2 != 3\)'
        ):
            self.assertStructuredAlmostEqual(a, b)
        self.assertStructuredAlmostEqual(a, b, allow_second_superset=True)
        a.append(3)

        self.assertStructuredAlmostEqual(a, b)
        b[1] -= 1.999e-7
        self.assertStructuredAlmostEqual(a, b)
        b[1] -= 1.999e-7
        with self.assertRaisesRegex(self.failureException, '2 !~= 1.999'):
            self.assertStructuredAlmostEqual(a, b)

    def test_assertStructuredAlmostEqual_dict(self):
        a = {1: 2, 3: 4}
        b = {1: 2, 3: 4, 5: 6}
        with self.assertRaisesRegex(
            self.failureException, r'mappings are different sizes \(2 != 3\)'
        ):
            self.assertStructuredAlmostEqual(a, b)
        self.assertStructuredAlmostEqual(a, b, allow_second_superset=True)
        a[5] = 6

        self.assertStructuredAlmostEqual(a, b)
        b[1] -= 1.999e-7
        self.assertStructuredAlmostEqual(a, b)
        b[1] -= 1.999e-7
        with self.assertRaisesRegex(self.failureException, '2 !~= 1.999'):
            self.assertStructuredAlmostEqual(a, b)

        del b[1]
        b[6] = 6
        with self.assertRaisesRegex(
            self.failureException, r'key \(1\) from first not found in second'
        ):
            self.assertStructuredAlmostEqual(a, b)

    def test_assertStructuredAlmostEqual_nested(self):
        a = {1.1: [1, 2, 3], 'a': 'hi', 3: {1: 2, 3: 4}}
        b = {1.1: [1, 2, 3], 'a': 'hi', 3: {1: 2, 3: 4}}
        self.assertStructuredAlmostEqual(a, b)
        b[1.1][2] -= 1.999e-7
        b[3][1] -= 9.999e-8
        self.assertStructuredAlmostEqual(a, b)
        b[1.1][2] -= 1.999e-7
        with self.assertRaisesRegex(self.failureException, '3 !~= 2.999'):
            self.assertStructuredAlmostEqual(a, b)

    def test_assertStructuredAlmostEqual_numericvalue(self):
        m = ConcreteModel()
        m.v = Var(initialize=2.0)
        m.p = Param(initialize=2.0)
        a = {1.1: [1, m.p, 3], 'a': 'hi', 3: {1: 2, 3: 4}}
        b = {1.1: [1, m.v, 3], 'a': 'hi', 3: {1: 2, 3: 4}}
        self.assertStructuredAlmostEqual(a, b)
        m.v.set_value(m.v.value - 1.999e-7)
        self.assertStructuredAlmostEqual(a, b)
        m.v.set_value(m.v.value - 1.999e-7)
        with self.assertRaisesRegex(self.failureException, '2.0 !~= 1.999'):
            self.assertStructuredAlmostEqual(a, b)

    def test_timeout_fcn_call(self):
        self.assertEqual(short_sleep(), 42)
        with self.assertRaisesRegex(TimeoutError, 'test timed out after 0.01 seconds'):
            long_sleep()
        with self.assertRaisesRegex(
            NameError, r"name 'foo' is not defined\s+Original traceback:"
        ):
            raise_exception()
        with self.assertRaisesRegex(AssertionError, r"^0 != 1$"):
            fail()

    @unittest.timeout(10)
    def test_timeout(self):
        self.assertEqual(0, 0)

    @unittest.expectedFailure
    @unittest.timeout(0.01)
    def test_timeout_timeout(self):
        time.sleep(1)
        self.assertEqual(0, 0)

    def timeout_skip(self, skip):
        if skip:
            self.skipTest("Skipping this test")
        self.assertEqual(0, 1)

    @unittest.timeout(10)
    def test_timeout_skip(self):
        self.timeout_skip(True)

    @unittest.timeout(10)
    def test_timeout_skip_pass(self):
        with self.assertRaisesRegex(AssertionError, r"0 != 1"):
            self.timeout_skip(False)

    @unittest.expectedFailure
    @unittest.timeout(10)
    def test_timeout_skip_fail(self):
        self.timeout_skip(False)

    @unittest.timeout(10)
    def bound_function(self):
        self.assertEqual(0, 0)

    def test_bound_function(self):
        if multiprocessing.get_start_method() == 'fork':
            self.bound_function()
            return
        with LoggingIntercept() as LOG:
            with self.assertRaises(
                (TypeError, EOFError, AttributeError, pickle.PicklingError)
            ):
                self.bound_function()
        self.assertIn("platform that does not support 'fork'", LOG.getvalue())
        self.assertIn("one of its arguments is not serializable", LOG.getvalue())

    @unittest.timeout(10, require_fork=True)
    def bound_function_require_fork(self):
        self.assertEqual(0, 0)

    def test_bound_function_require_fork(self):
        if multiprocessing.get_start_method() == 'fork':
            self.bound_function_require_fork()
            return
        with self.assertRaisesRegex(
            unittest.SkipTest, r"timeout\(\) requires unavailable fork interface"
        ):
            self.bound_function_require_fork()


baseline = """
[    0.00] Setting up Pyomo environment
[    0.00] Applying Pyomo preprocessing actions
[    0.00] Creating model
[    0.00] Applying solver
[    0.05] Processing results
    Number of solutions: 1
    Solution Information
      Gap: None
      Status: optimal
      Function Value: -9.99943939749e-05
    Solver results file: results.yml
[    0.05] Applying Pyomo postprocessing actions
[    0.05] Pyomo Finished
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Problem Information
# ----------------------------------------------------------
Problem:
- Lower bound: -inf
  Upper bound: inf
  Number of objectives: 1
  Number of constraints: 3
  Number of variables: 3
  Sense: unknown
# ----------------------------------------------------------
#   Solver Information
# ----------------------------------------------------------
Solver:
- Status: ok
  Message: Ipopt 3.12.3\x3a Optimal Solution Found
  Termination condition: optimal
  Id: 0
  Error rc: 0
  Time: 0.0408430099487
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution:
- number of solutions: 1
  number of solutions displayed: 1
- Gap: None
  Status: optimal
  Message: Ipopt 3.12.3\x3a Optimal Solution Found
  Objective:
    f1:
      Value: -9.99943939749e-05
  Variable:
    compl.v:
      Value: 9.99943939749e-05
    y:
      Value: 9.99943939749e-05
  Constraint: No values
"""

pass_ref = """
[    0.00] Setting up Pyomo environment
[    0.00] Applying Pyomo preprocessing actions
WARNING: DEPRECATED: The Model.preprocess() method is deprecated and no longer
    performs any actions  (deprecated in 6.0) (called from <stdin>:1)
[    0.00] Creating model
[    0.01] Applying solver
[    0.06] Processing results
    Number of solutions: 1
    Solution Information
      Gap: None
      Status: optimal
      Function Value: -0.00010001318188373491
    Solver results file: results.yml
[    0.06] Applying Pyomo postprocessing actions
[    0.06] Pyomo Finished
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Problem Information
# ----------------------------------------------------------
Problem:
- Lower bound: -inf
  Upper bound: inf
  Number of objectives: 1
  Number of constraints: 3
  Number of variables: 3
  Sense: unknown
# ----------------------------------------------------------
#   Solver Information
# ----------------------------------------------------------
Solver:
- Status: ok
  Message: Ipopt 3.14.13\x3a Optimal Solution Found
  Termination condition: optimal
  Id: 0
  Error rc: 0
  Time: 0.04224729537963867
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution:
- number of solutions: 1
  number of solutions displayed: 1
- Gap: None
  Status: optimal
  Message: Ipopt 3.14.13\x3a Optimal Solution Found
  Objective:
    f1:
      Value: -0.00010001318188373491
  Variable:
    compl.v:
      Value: 9.99943939749205e-05
    x:
      Value: -9.39395440720558e-09
    y:
      Value: 9.99943939749205e-05
  Constraint: No values

"""

fail_ref = """
[    0.00] Setting up Pyomo environment
[    0.00] Applying Pyomo preprocessing actions
[    0.00] Creating model
[    0.01] Applying solver
[    0.06] Processing results
    Number of solutions: 1
    Solution Information
      Gap: None
      Status: optimal
      Function Value: -0.00010001318188373491
    Solver results file: results.yml
[    0.06] Applying Pyomo postprocessing actions
[    0.06] Pyomo Finished
# ==========================================================
# = Solver Results                                         =
# ==========================================================
# ----------------------------------------------------------
#   Problem Information
# ----------------------------------------------------------
Problem:
- Lower bound: -inf
  Upper bound: inf
  Number of objectives: 1
  Number of constraints: 3
  Number of variables: 3
  Sense: unknown
# ----------------------------------------------------------
#   Solver Information
# ----------------------------------------------------------
Solver:
- Status: ok
  Message: Ipopt 3.14.13\x3a Optimal Solution Found
  Termination condition: optimal
  Id: 0
  Error rc: 0
  Time: 0.04224729537963867
# ----------------------------------------------------------
#   Solution Information
# ----------------------------------------------------------
Solution:
- number of solutions: 1
  number of solutions displayed: 1
- Gap: None
  Status: optimal
  Message: Ipopt 3.14.13\x3a Optimal Solution Found
  Objective:
    f1:
      Value: -0.00010001318188373491
  Variable:
    compl.v:
      Value: 9.79943939749205e-05
    x:
      Value: -9.39395440720558e-09
    y:
      Value: 9.99943939749205e-05
  Constraint: No values

"""


class TestBaselineTestDriver(unittest.BaselineTestDriver, unittest.TestCase):
    solver_dependencies = {}
    package_dependencies = {}

    def test_baseline_pass(self):
        self.compare_baseline(pass_ref, baseline, abstol=1e-6)

        with self.assertRaises(self.failureException):
            with capture_output() as OUT:
                self.compare_baseline(pass_ref, baseline, None)
        self.assertEqual(
            OUT.getvalue(),
            f"""---------------------------------
BASELINE FILE
---------------------------------
{baseline}
=================================
---------------------------------
TEST OUTPUT FILE
---------------------------------
{pass_ref}
""",
        )

    def test_baseline_fail(self):
        with self.assertRaises(self.failureException):
            with capture_output() as OUT:
                self.compare_baseline(fail_ref, baseline)
        self.assertEqual(
            OUT.getvalue(),
            f"""---------------------------------
BASELINE FILE
---------------------------------
{baseline}
=================================
---------------------------------
TEST OUTPUT FILE
---------------------------------
{fail_ref}
""",
        )

    def test_testcase_collection(self):
        with TempfileManager.new_context() as TMP:
            tmpdir = TMP.create_tempdir()
            for fname in (
                'a.py',
                'b.py',
                'b.txt',
                'c.py',
                'c.sh',
                'c.yml',
                'd.sh',
                'd.txt',
                'e.sh',
            ):
                with open(os.path.join(tmpdir, fname), 'w'):
                    pass

            py_tests, sh_tests = unittest.BaselineTestDriver.gather_tests([tmpdir])
            self.assertEqual(
                py_tests,
                [
                    (
                        os.path.basename(tmpdir) + '_b',
                        os.path.join(tmpdir, 'b.py'),
                        os.path.join(tmpdir, 'b.txt'),
                    )
                ],
            )
            self.assertEqual(
                sh_tests,
                [
                    (
                        os.path.basename(tmpdir) + '_c',
                        os.path.join(tmpdir, 'c.sh'),
                        os.path.join(tmpdir, 'c.yml'),
                    ),
                    (
                        os.path.basename(tmpdir) + '_d',
                        os.path.join(tmpdir, 'd.sh'),
                        os.path.join(tmpdir, 'd.txt'),
                    ),
                ],
            )

            self.python_test_driver(*py_tests[0])

            _update_baselines = os.environ.pop('PYOMO_TEST_UPDATE_BASELINES', None)
            try:
                with open(os.path.join(tmpdir, 'b.py'), 'w') as FILE:
                    FILE.write('print("Hello, World")\n')

                with self.assertRaises(self.failureException):
                    self.python_test_driver(*py_tests[0])
                with open(os.path.join(tmpdir, 'b.txt'), 'r') as FILE:
                    self.assertEqual(FILE.read(), "")

                os.environ['PYOMO_TEST_UPDATE_BASELINES'] = '1'

                with self.assertRaises(self.failureException):
                    self.python_test_driver(*py_tests[0])
                with open(os.path.join(tmpdir, 'b.txt'), 'r') as FILE:
                    self.assertEqual(FILE.read(), "Hello, World\n")

            finally:
                os.environ.pop('PYOMO_TEST_UPDATE_BASELINES', None)
                if _update_baselines is not None:
                    os.environ['PYOMO_TEST_UPDATE_BASELINES'] = _update_baselines

            self.shell_test_driver(*sh_tests[1])
            _update_baselines = os.environ.pop('PYOMO_TEST_UPDATE_BASELINES', None)
            try:
                with open(os.path.join(tmpdir, 'd.sh'), 'w') as FILE:
                    FILE.write('echo "Hello, World"\n')

                with self.assertRaises(self.failureException):
                    self.shell_test_driver(*sh_tests[1])
                with open(os.path.join(tmpdir, 'd.txt'), 'r') as FILE:
                    self.assertEqual(FILE.read(), "")

                os.environ['PYOMO_TEST_UPDATE_BASELINES'] = '1'

                with self.assertRaises(self.failureException):
                    self.shell_test_driver(*sh_tests[1])
                with open(os.path.join(tmpdir, 'd.txt'), 'r') as FILE:
                    self.assertEqual(FILE.read(), "Hello, World\n")

            finally:
                os.environ.pop('PYOMO_TEST_UPDATE_BASELINES', None)
                if _update_baselines is not None:
                    os.environ['PYOMO_TEST_UPDATE_BASELINES'] = _update_baselines


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import Factory
from pyomo.common.collections import ComponentSet
from pyomo.common.errors import MouseTrap
from pyomo.common.deprecation import deprecated
from pyomo.common.modeling import unique_component_name
from pyomo.common.timing import TransformationTimer


class TransformationInfo:
    pass


class TransformationData:
    """
    This is a container class that supports named data objects.
    """

    def __init__(self):
        self._data = {}

    def __getitem__(self, name):
        if not name in self._data:
            self._data[name] = TransformationInfo()
        return self._data[name]


class Transformation:
    """
    Base class for all model transformations.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "transformation")
        # super(Transformation, self).__init__(**kwds)

    #
    # Support "with" statements.
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass

    @deprecated(
        "Transformation.apply() has been deprecated.  Please use either "
        "Transformation.apply_to() for in-place transformations or "
        "Transformation.create_using() for transformations that create a "
        "new, independent transformed model instance.",
        version='4.3.11323',
    )
    def apply(self, model, **kwds):
        inplace = kwds.pop('inplace', True)
        if inplace:
            self.apply_to(model, **kwds)
        else:
            return self.create_using(model, **kwds)

    def apply_to(self, model, **kwds):
        """
        Apply the transformation to the given model.
        """
        timer = TransformationTimer(self, 'in-place')
        if not hasattr(model, '_transformation_data'):
            model._transformation_data = TransformationData()
        reverse_token = self._apply_to(model, **kwds)
        timer.report()

        return reverse_token

    def create_using(self, model, **kwds):
        """
        Create a new model with this transformation
        """
        timer = TransformationTimer(self, 'out-of-place')
        if not hasattr(model, '_transformation_data'):
            model._transformation_data = TransformationData()
        new_model = self._create_using(model, **kwds)
        timer.report()
        return new_model

    def _apply_to(self, model, **kwds):
        raise RuntimeError("The Transformation.apply_to method is not implemented.")

    def _create_using(self, model, **kwds):
        # Put all the kwds onto the model so that when we clone the
        # model any references to things on the model are correctly
        # updated to point to the new instance.  Note that users &
        # transformation developers cannot rely on things happening by
        # argument side effect.
        name = unique_component_name(model, '_kwds')
        setattr(model, name, kwds)
        instance = model.clone()
        kwds = getattr(instance, name)
        delattr(model, name)
        delattr(instance, name)
        self._apply_to(instance, **kwds)
        return instance


class ReverseTransformationToken:
    """
    Class returned by reversible transformations' apply_to methods that
    can be passed back to the transformation in order to revert its changes
    to the model.

    We store the transformation that created it, so that we have some basic
    error checking when the user attempts to revert, and we store a dictionary
    that can be whatever the transformation wants/needs in order to revert
    itself.

    args:
        transformation: The class of the transformation that created this token
        model: The model being transformed when this token was created
        targets: The targets on 'model' being transformed when this token
                 was created.
        reverse_dict: Dictionary with everything the transformation needs to
                      undo itself.
    """

    def __init__(self, transformation, model, targets, reverse_dict):
        self._transformation = transformation
        self._model = model
        self._targets = ComponentSet(targets)
        self._reverse_dict = reverse_dict

    @property
    def transformation(self):
        return self._transformation

    @property
    def reverse_dict(self):
        return self._reverse_dict

    def check_token_valid(self, cls, model, targets):
        if cls is not self._transformation:
            raise ValueError(
                "Attempting to reverse transformation of class '%s' "
                "using a token created by a transformation of class "
                "'%s'. Cannot revert transformation with a token from "
                "another transformation." % (cls, self._transformation)
            )
        if model is not self._model:
            raise MouseTrap(
                "A reverse transformation was called on model '%s', but the "
                "transformation that created this token was created from "
                "model '%s'. We do not currently support reversing "
                "transformations on clones of the transformed model."
                % (model.name, self._model.name)
            )
        # TODO: Do we need to pass targets into this? I'm thinking no because
        # people can untransform selectively. We just need to enforce that the
        # targets given here were indeed transformed, but they don't have to
        # correspond exactly to what happened before. I think we leave that for
        # the transformation to sort out


TransformationFactory = Factory('transformation type')


@deprecated(version='4.3.11323')
def apply_transformation(*args, **kwds):
    if len(args) == 0:
        return list(TransformationFactory)
    xfrm = TransformationFactory(args[0])
    if len(args) == 1 or xfrm is None:
        return xfrm
    tmp = (args[1],)
    return xfrm.apply(*tmp, **kwds)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.common.dependencies import attempt_import
import pyomo.common.unittest as unittest

parameterized, param_available = attempt_import('parameterized')
parameterized = parameterized.parameterized
from pyomo.contrib.appsi.base import TerminationCondition, Results, PersistentSolver
from pyomo.contrib.appsi.cmodel import cmodel_available
from pyomo.contrib.appsi.solvers import Gurobi, Ipopt, Cplex, Cbc, Highs, MAiNGO
from typing import Type
from pyomo.core.expr.numeric_expr import LinearExpression
import os

numpy, numpy_available = attempt_import('numpy')
import random
from pyomo import gdp

if not param_available:
    raise unittest.SkipTest('Parameterized is not available.')

all_solvers = [
    ('gurobi', Gurobi),
    ('ipopt', Ipopt),
    ('cplex', Cplex),
    ('cbc', Cbc),
    ('highs', Highs),
    ('maingo', MAiNGO),
]
mip_solvers = [
    ('gurobi', Gurobi),
    ('cplex', Cplex),
    ('cbc', Cbc),
    ('highs', Highs),
    ('maingo', MAiNGO),
]
nlp_solvers = [('ipopt', Ipopt), ('maingo', MAiNGO)]
qcp_solvers = [
    ('gurobi', Gurobi),
    ('ipopt', Ipopt),
    ('cplex', Cplex),
    ('maingo', MAiNGO),
]
miqcqp_solvers = [('gurobi', Gurobi), ('cplex', Cplex), ('maingo', MAiNGO)]
only_child_vars_options = [True, False]


"""
The tests in this file are used to ensure basic functionality/API works with all solvers

Feature                                    Tested
-------                                    ------
config time_limit                          
config tee                                 
config load_solution True                  x
config load_solution False                 x     
results termination condition optimal      x
results termination condition infeasible   x
load_vars                                  
get_duals                                  x
get_reduced_costs                          x
range constraints                          x
MILP
Model updates - added constraints          x
Model updates - removed constraints        x
Model updates - added vars
Model updates - removed vars
Model updates - changed named expression
Model updates - mutable param modified     x
Model updates - var modified
Model updates - objective changed
Model updates - constraint modified
No objective
No constraints                             x
bounds                                     x
best feasible objective                    x
best objective bound                       x
fixed variables
"""


def _load_tests(solver_list, only_child_vars_list):
    res = list()
    for solver_name, solver in solver_list:
        for child_var_option in only_child_vars_list:
            test_name = f"{solver_name}_only_child_vars_{child_var_option}"
            res.append((test_name, solver, child_var_option))
    return res


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
@unittest.skipUnless(numpy_available, 'numpy is not available')
class TestSolvers(unittest.TestCase):
    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_remove_variable_and_objective(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        # this test is for issue #2888
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(2, None))
        m.obj = pyo.Objective(expr=m.x)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 2)

        del m.x
        del m.obj
        m.x = pyo.Var(bounds=(2, None))
        m.obj = pyo.Objective(expr=m.x)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 2)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_stale_vars(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= m.x)
        m.c2 = pyo.Constraint(expr=m.y >= -m.x)
        m.x.value = 1
        m.y.value = 1
        m.z.value = 1
        self.assertFalse(m.x.stale)
        self.assertFalse(m.y.stale)
        self.assertFalse(m.z.stale)

        res = opt.solve(m)
        self.assertFalse(m.x.stale)
        self.assertFalse(m.y.stale)
        self.assertTrue(m.z.stale)

        opt.config.load_solution = False
        res = opt.solve(m)
        self.assertTrue(m.x.stale)
        self.assertTrue(m.y.stale)
        self.assertTrue(m.z.stale)
        res.solution_loader.load_vars()
        self.assertFalse(m.x.stale)
        self.assertFalse(m.y.stale)
        self.assertTrue(m.z.stale)

        res = opt.solve(m)
        self.assertTrue(m.x.stale)
        self.assertTrue(m.y.stale)
        self.assertTrue(m.z.stale)
        res.solution_loader.load_vars([m.y])
        self.assertTrue(m.x.stale)
        self.assertFalse(m.y.stale)
        self.assertTrue(m.z.stale)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_range_constraint(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.obj = pyo.Objective(expr=m.x)
        m.c = pyo.Constraint(expr=(-1, m.x, 1))
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, -1)
        if opt_class != MAiNGO:
            duals = opt.get_duals()
            self.assertAlmostEqual(duals[m.c], 1)
        m.obj.sense = pyo.maximize
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 1)
        if opt_class != MAiNGO:
            duals = opt.get_duals()
            self.assertAlmostEqual(duals[m.c], 1)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_reduced_costs(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-1, 1))
        m.y = pyo.Var(bounds=(-2, 2))
        m.obj = pyo.Objective(expr=3 * m.x + 4 * m.y)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, -1)
        self.assertAlmostEqual(m.y.value, -2)
        if opt_class != MAiNGO:
            rc = opt.get_reduced_costs()
            self.assertAlmostEqual(rc[m.x], 3)
            self.assertAlmostEqual(rc[m.y], 4)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_reduced_costs2(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-1, 1))
        m.obj = pyo.Objective(expr=m.x)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, -1)
        if opt_class != MAiNGO:
            rc = opt.get_reduced_costs()
            self.assertAlmostEqual(rc[m.x], 1)
        m.obj.sense = pyo.maximize
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 1)
        if opt_class != MAiNGO:
            rc = opt.get_reduced_costs()
            self.assertAlmostEqual(rc[m.x], 1)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_param_changes(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a1 = pyo.Param(mutable=True)
        m.a2 = pyo.Param(mutable=True)
        m.b1 = pyo.Param(mutable=True)
        m.b2 = pyo.Param(mutable=True)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=(0, m.y - m.a1 * m.x - m.b1, None))
        m.c2 = pyo.Constraint(expr=(None, -m.y + m.a2 * m.x + m.b2, 0))

        params_to_test = [(1, -1, 2, 1), (1, -2, 2, 1), (1, -1, 3, 1)]
        for a1, a2, b1, b2 in params_to_test:
            m.a1.value = a1
            m.a2.value = a2
            m.b1.value = b1
            m.b2.value = b2
            res: Results = opt.solve(m)
            self.assertEqual(res.termination_condition, TerminationCondition.optimal)
            self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
            self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
            self.assertAlmostEqual(res.best_feasible_objective, m.y.value)
            self.assertTrue(res.best_objective_bound <= m.y.value)
            if opt_class != MAiNGO:
                duals = opt.get_duals()
                self.assertAlmostEqual(duals[m.c1], (1 + a1 / (a2 - a1)))
                self.assertAlmostEqual(duals[m.c2], a1 / (a2 - a1))

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_immutable_param(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        """
        This test is important because component_data_objects returns immutable params as floats.
        We want to make sure we process these correctly.
        """
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a1 = pyo.Param(mutable=True)
        m.a2 = pyo.Param(initialize=-1)
        m.b1 = pyo.Param(mutable=True)
        m.b2 = pyo.Param(mutable=True)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=(0, m.y - m.a1 * m.x - m.b1, None))
        m.c2 = pyo.Constraint(expr=(None, -m.y + m.a2 * m.x + m.b2, 0))

        params_to_test = [(1, 2, 1), (1, 2, 1), (1, 3, 1)]
        for a1, b1, b2 in params_to_test:
            a2 = m.a2.value
            m.a1.value = a1
            m.b1.value = b1
            m.b2.value = b2
            res: Results = opt.solve(m)
            self.assertEqual(res.termination_condition, TerminationCondition.optimal)
            self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
            self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
            self.assertAlmostEqual(res.best_feasible_objective, m.y.value)
            self.assertTrue(res.best_objective_bound <= m.y.value)
            if opt_class != MAiNGO:
                duals = opt.get_duals()
                self.assertAlmostEqual(duals[m.c1], (1 + a1 / (a2 - a1)))
                self.assertAlmostEqual(duals[m.c2], a1 / (a2 - a1))

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_equality(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a1 = pyo.Param(mutable=True)
        m.a2 = pyo.Param(mutable=True)
        m.b1 = pyo.Param(mutable=True)
        m.b2 = pyo.Param(mutable=True)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y == m.a1 * m.x + m.b1)
        m.c2 = pyo.Constraint(expr=m.y == m.a2 * m.x + m.b2)

        params_to_test = [(1, -1, 2, 1), (1, -2, 2, 1), (1, -1, 3, 1)]
        for a1, a2, b1, b2 in params_to_test:
            m.a1.value = a1
            m.a2.value = a2
            m.b1.value = b1
            m.b2.value = b2
            res: Results = opt.solve(m)
            self.assertEqual(res.termination_condition, TerminationCondition.optimal)
            self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
            self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
            self.assertAlmostEqual(res.best_feasible_objective, m.y.value)
            self.assertTrue(res.best_objective_bound <= m.y.value)
            if opt_class != MAiNGO:
                duals = opt.get_duals()
                self.assertAlmostEqual(duals[m.c1], (1 + a1 / (a2 - a1)))
                self.assertAlmostEqual(duals[m.c2], -a1 / (a2 - a1))

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_linear_expression(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a1 = pyo.Param(mutable=True)
        m.a2 = pyo.Param(mutable=True)
        m.b1 = pyo.Param(mutable=True)
        m.b2 = pyo.Param(mutable=True)
        m.obj = pyo.Objective(expr=m.y)
        e = LinearExpression(
            constant=m.b1, linear_coefs=[-1, m.a1], linear_vars=[m.y, m.x]
        )
        m.c1 = pyo.Constraint(expr=e == 0)
        e = LinearExpression(
            constant=m.b2, linear_coefs=[-1, m.a2], linear_vars=[m.y, m.x]
        )
        m.c2 = pyo.Constraint(expr=e == 0)

        params_to_test = [(1, -1, 2, 1), (1, -2, 2, 1), (1, -1, 3, 1)]
        for a1, a2, b1, b2 in params_to_test:
            m.a1.value = a1
            m.a2.value = a2
            m.b1.value = b1
            m.b2.value = b2
            res: Results = opt.solve(m)
            self.assertEqual(res.termination_condition, TerminationCondition.optimal)
            self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
            self.assertAlmostEqual(res.best_feasible_objective, m.y.value)
            self.assertTrue(res.best_objective_bound <= m.y.value)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_no_objective(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a1 = pyo.Param(mutable=True)
        m.a2 = pyo.Param(mutable=True)
        m.b1 = pyo.Param(mutable=True)
        m.b2 = pyo.Param(mutable=True)
        m.c1 = pyo.Constraint(expr=m.y == m.a1 * m.x + m.b1)
        m.c2 = pyo.Constraint(expr=m.y == m.a2 * m.x + m.b2)
        opt.config.stream_solver = True

        params_to_test = [(1, -1, 2, 1), (1, -2, 2, 1), (1, -1, 3, 1)]
        for a1, a2, b1, b2 in params_to_test:
            m.a1.value = a1
            m.a2.value = a2
            m.b1.value = b1
            m.b2.value = b2
            res: Results = opt.solve(m)
            self.assertEqual(res.termination_condition, TerminationCondition.optimal)
            self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
            self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
            self.assertEqual(res.best_feasible_objective, None)
            self.assertEqual(res.best_objective_bound, None)
            if opt_class != MAiNGO:
                duals = opt.get_duals()
                self.assertAlmostEqual(duals[m.c1], 0)
                self.assertAlmostEqual(duals[m.c2], 0)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_add_remove_cons(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        a1 = -1
        a2 = 1
        b1 = 1
        b2 = 2
        a3 = 1
        b3 = 3
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= a1 * m.x + b1)
        m.c2 = pyo.Constraint(expr=m.y >= a2 * m.x + b2)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
        self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
        self.assertAlmostEqual(res.best_feasible_objective, m.y.value)
        self.assertTrue(res.best_objective_bound <= m.y.value)
        if opt_class != MAiNGO:
            duals = opt.get_duals()
            self.assertAlmostEqual(duals[m.c1], -(1 + a1 / (a2 - a1)))
            self.assertAlmostEqual(duals[m.c2], a1 / (a2 - a1))

        m.c3 = pyo.Constraint(expr=m.y >= a3 * m.x + b3)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, (b3 - b1) / (a1 - a3))
        self.assertAlmostEqual(m.y.value, a1 * (b3 - b1) / (a1 - a3) + b1)
        self.assertAlmostEqual(res.best_feasible_objective, m.y.value)
        self.assertTrue(res.best_objective_bound <= m.y.value)
        if opt_class != MAiNGO:
            duals = opt.get_duals()
            self.assertAlmostEqual(duals[m.c1], -(1 + a1 / (a3 - a1)))
            self.assertAlmostEqual(duals[m.c2], 0)
            self.assertAlmostEqual(duals[m.c3], a1 / (a3 - a1))

        del m.c3
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
        self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
        self.assertAlmostEqual(res.best_feasible_objective, m.y.value)
        self.assertTrue(res.best_objective_bound <= m.y.value)
        if opt_class != MAiNGO:
            duals = opt.get_duals()
            self.assertAlmostEqual(duals[m.c1], -(1 + a1 / (a2 - a1)))
            self.assertAlmostEqual(duals[m.c2], a1 / (a2 - a1))

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_results_infeasible(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= m.x)
        m.c2 = pyo.Constraint(expr=m.y <= m.x - 1)
        with self.assertRaises(Exception):
            res = opt.solve(m)
        opt.config.load_solution = False
        res = opt.solve(m)
        self.assertNotEqual(res.termination_condition, TerminationCondition.optimal)
        if opt_class is Ipopt:
            acceptable_termination_conditions = {
                TerminationCondition.infeasible,
                TerminationCondition.unbounded,
            }
        else:
            acceptable_termination_conditions = {
                TerminationCondition.infeasible,
                TerminationCondition.infeasibleOrUnbounded,
            }
        self.assertIn(res.termination_condition, acceptable_termination_conditions)
        self.assertAlmostEqual(m.x.value, None)
        self.assertAlmostEqual(m.y.value, None)
        self.assertTrue(res.best_feasible_objective is None)

        with self.assertRaisesRegex(
            RuntimeError, '.*does not currently have a valid solution.*'
        ):
            res.solution_loader.load_vars()
        if opt_class != MAiNGO:
            with self.assertRaisesRegex(
                RuntimeError, '.*does not currently have valid duals.*'
            ):
                res.solution_loader.get_duals()
            with self.assertRaisesRegex(
                RuntimeError, '.*does not currently have valid reduced costs.*'
            ):
                res.solution_loader.get_reduced_costs()

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_duals(self, name: str, opt_class: Type[PersistentSolver], only_child_vars):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y - m.x >= 0)
        m.c2 = pyo.Constraint(expr=m.y + m.x - 2 >= 0)

        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 1)
        if opt_class != MAiNGO:
            duals = opt.get_duals()
            self.assertAlmostEqual(duals[m.c1], 0.5)
            self.assertAlmostEqual(duals[m.c2], 0.5)

            duals = opt.get_duals(cons_to_load=[m.c1])
            self.assertAlmostEqual(duals[m.c1], 0.5)
            self.assertNotIn(m.c2, duals)

    @parameterized.expand(input=_load_tests(qcp_solvers, only_child_vars_options))
    def test_mutable_quadratic_coefficient(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a = pyo.Param(initialize=1, mutable=True)
        m.b = pyo.Param(initialize=-1, mutable=True)
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c = pyo.Constraint(expr=m.y >= (m.a * m.x + m.b) ** 2)

        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0.41024548525899274, 4)
        self.assertAlmostEqual(m.y.value, 0.34781038127030117, 4)
        m.a.value = 2
        m.b.value = -0.5
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0.10256137418973625, 4)
        self.assertAlmostEqual(m.y.value, 0.0869525991355825, 4)

    @parameterized.expand(input=_load_tests(qcp_solvers, only_child_vars_options))
    def test_mutable_quadratic_objective(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a = pyo.Param(initialize=1, mutable=True)
        m.b = pyo.Param(initialize=-1, mutable=True)
        m.c = pyo.Param(initialize=1, mutable=True)
        m.d = pyo.Param(initialize=1, mutable=True)
        m.obj = pyo.Objective(expr=m.x**2 + m.c * m.y**2 + m.d * m.x)
        m.ccon = pyo.Constraint(expr=m.y >= (m.a * m.x + m.b) ** 2)

        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0.2719178742733325, 4)
        self.assertAlmostEqual(m.y.value, 0.5301035741688002, 4)
        m.c.value = 3.5
        m.d.value = -1
        res = opt.solve(m)

        self.assertAlmostEqual(m.x.value, 0.6962249634573562, 4)
        self.assertAlmostEqual(m.y.value, 0.09227926676152151, 4)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_fixed_vars(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        for treat_fixed_vars_as_params in [True, False]:
            opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
            opt.update_config.treat_fixed_vars_as_params = treat_fixed_vars_as_params
            if not opt.available():
                raise unittest.SkipTest
            m = pyo.ConcreteModel()
            m.x = pyo.Var()
            m.x.fix(0)
            m.y = pyo.Var()
            a1 = 1
            a2 = -1
            b1 = 1
            b2 = 2
            m.obj = pyo.Objective(expr=m.y)
            m.c1 = pyo.Constraint(expr=m.y >= a1 * m.x + b1)
            m.c2 = pyo.Constraint(expr=m.y >= a2 * m.x + b2)
            res = opt.solve(m)
            self.assertAlmostEqual(m.x.value, 0)
            self.assertAlmostEqual(m.y.value, 2)
            m.x.unfix()
            res = opt.solve(m)
            self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
            self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
            m.x.fix(0)
            res = opt.solve(m)
            self.assertAlmostEqual(m.x.value, 0)
            self.assertAlmostEqual(m.y.value, 2)
            m.x.value = 2
            res = opt.solve(m)
            self.assertAlmostEqual(m.x.value, 2)
            self.assertAlmostEqual(m.y.value, 3)
            m.x.value = 0
            res = opt.solve(m)
            self.assertAlmostEqual(m.x.value, 0)
            self.assertAlmostEqual(m.y.value, 2)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_fixed_vars_2(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        opt.update_config.treat_fixed_vars_as_params = True
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.x.fix(0)
        m.y = pyo.Var()
        a1 = 1
        a2 = -1
        b1 = 1
        b2 = 2
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= a1 * m.x + b1)
        m.c2 = pyo.Constraint(expr=m.y >= a2 * m.x + b2)
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 2)
        m.x.unfix()
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
        self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
        m.x.fix(0)
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 2)
        m.x.value = 2
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 2)
        self.assertAlmostEqual(m.y.value, 3)
        m.x.value = 0
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 2)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_fixed_vars_3(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        opt.update_config.treat_fixed_vars_as_params = True
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y)
        m.c1 = pyo.Constraint(expr=m.x == 2 / m.y)
        m.y.fix(1)
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 2)

    @parameterized.expand(input=_load_tests(nlp_solvers, only_child_vars_options))
    def test_fixed_vars_4(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        opt.update_config.treat_fixed_vars_as_params = True
        if not opt.available() or opt_class == MAiNGO:
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.x == 2 / m.y)
        m.y.fix(1)
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 2)
        m.y.unfix()
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 2**0.5)
        self.assertAlmostEqual(m.y.value, 2**0.5)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_mutable_param_with_range(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        try:
            import numpy as np
        except:
            raise unittest.SkipTest('numpy is not available')
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a1 = pyo.Param(initialize=0, mutable=True)
        m.a2 = pyo.Param(initialize=0, mutable=True)
        m.b1 = pyo.Param(initialize=0, mutable=True)
        m.b2 = pyo.Param(initialize=0, mutable=True)
        m.c1 = pyo.Param(initialize=0, mutable=True)
        m.c2 = pyo.Param(initialize=0, mutable=True)
        m.obj = pyo.Objective(expr=m.y)
        m.con1 = pyo.Constraint(expr=(m.b1, m.y - m.a1 * m.x, m.c1))
        m.con2 = pyo.Constraint(expr=(m.b2, m.y - m.a2 * m.x, m.c2))

        np.random.seed(0)
        params_to_test = [
            (
                np.random.uniform(0, 10),
                np.random.uniform(-10, 0),
                np.random.uniform(-5, 2.5),
                np.random.uniform(-5, 2.5),
                np.random.uniform(2.5, 10),
                np.random.uniform(2.5, 10),
                pyo.minimize,
            ),
            (
                np.random.uniform(0, 10),
                np.random.uniform(-10, 0),
                np.random.uniform(-5, 2.5),
                np.random.uniform(-5, 2.5),
                np.random.uniform(2.5, 10),
                np.random.uniform(2.5, 10),
                pyo.maximize,
            ),
            (
                np.random.uniform(0, 10),
                np.random.uniform(-10, 0),
                np.random.uniform(-5, 2.5),
                np.random.uniform(-5, 2.5),
                np.random.uniform(2.5, 10),
                np.random.uniform(2.5, 10),
                pyo.minimize,
            ),
            (
                np.random.uniform(0, 10),
                np.random.uniform(-10, 0),
                np.random.uniform(-5, 2.5),
                np.random.uniform(-5, 2.5),
                np.random.uniform(2.5, 10),
                np.random.uniform(2.5, 10),
                pyo.maximize,
            ),
        ]
        for a1, a2, b1, b2, c1, c2, sense in params_to_test:
            m.a1.value = float(a1)
            m.a2.value = float(a2)
            m.b1.value = float(b1)
            m.b2.value = float(b2)
            m.c1.value = float(c1)
            m.c2.value = float(c2)
            m.obj.sense = sense
            res: Results = opt.solve(m)
            self.assertEqual(res.termination_condition, TerminationCondition.optimal)
            if sense is pyo.minimize:
                self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2), 6)
                self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1, 6)
                self.assertAlmostEqual(res.best_feasible_objective, m.y.value, 6)
                self.assertTrue(res.best_objective_bound <= m.y.value + 1e-12)
                if opt_class != MAiNGO:
                    duals = opt.get_duals()
                    self.assertAlmostEqual(duals[m.con1], (1 + a1 / (a2 - a1)), 6)
                    self.assertAlmostEqual(duals[m.con2], -a1 / (a2 - a1), 6)
            else:
                self.assertAlmostEqual(m.x.value, (c2 - c1) / (a1 - a2), 6)
                self.assertAlmostEqual(m.y.value, a1 * (c2 - c1) / (a1 - a2) + c1, 6)
                self.assertAlmostEqual(res.best_feasible_objective, m.y.value, 6)
                self.assertTrue(res.best_objective_bound >= m.y.value - 1e-12)
                if opt_class != MAiNGO:
                    duals = opt.get_duals()
                    self.assertAlmostEqual(duals[m.con1], (1 + a1 / (a2 - a1)), 6)
                    self.assertAlmostEqual(duals[m.con2], -a1 / (a2 - a1), 6)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_add_and_remove_vars(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.y = pyo.Var(bounds=(-1, None))
        m.obj = pyo.Objective(expr=m.y)
        opt.update_config.update_params = False
        opt.update_config.update_vars = False
        opt.update_config.update_constraints = False
        opt.update_config.update_named_expressions = False
        opt.update_config.check_for_new_or_removed_params = False
        opt.update_config.check_for_new_or_removed_constraints = False
        opt.update_config.check_for_new_or_removed_vars = False
        opt.config.load_solution = False
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        opt.load_vars()
        self.assertAlmostEqual(m.y.value, -1)
        m.x = pyo.Var()
        a1 = 1
        a2 = -1
        b1 = 2
        b2 = 1
        m.c1 = pyo.Constraint(expr=(0, m.y - a1 * m.x - b1, None))
        m.c2 = pyo.Constraint(expr=(None, -m.y + a2 * m.x + b2, 0))
        if only_child_vars:
            opt.add_variables([m.x])
        opt.add_constraints([m.c1, m.c2])
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        opt.load_vars()
        self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
        self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
        opt.remove_constraints([m.c1, m.c2])
        if only_child_vars:
            opt.remove_variables([m.x])
        m.x.value = None
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        opt.load_vars()
        self.assertEqual(m.x.value, None)
        self.assertAlmostEqual(m.y.value, -1)
        with self.assertRaises(Exception):
            opt.load_vars([m.x])

    @parameterized.expand(input=_load_tests(nlp_solvers, only_child_vars_options))
    def test_exp(self, name: str, opt_class: Type[PersistentSolver], only_child_vars):
        opt = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.y >= pyo.exp(m.x))
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, -0.42630274815985264, 6)
        self.assertAlmostEqual(m.y.value, 0.6529186341994245, 6)

    @parameterized.expand(input=_load_tests(nlp_solvers, only_child_vars_options))
    def test_log(self, name: str, opt_class: Type[PersistentSolver], only_child_vars):
        opt = opt_class(only_child_vars=only_child_vars)
        if not opt.available() or opt_class == MAiNGO:
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1)
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.y <= pyo.log(m.x))
        res = opt.solve(m)
        self.assertAlmostEqual(m.x.value, 0.6529186341994245)
        self.assertAlmostEqual(m.y.value, -0.42630274815985264)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_with_numpy(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.y)
        a1 = 1
        b1 = 3
        a2 = -2
        b2 = 1
        m.c1 = pyo.Constraint(
            expr=(numpy.float64(0), m.y - numpy.int64(1) * m.x - numpy.float32(3), None)
        )
        m.c2 = pyo.Constraint(
            expr=(
                None,
                -m.y + numpy.int32(-2) * m.x + numpy.float64(1),
                numpy.float16(0),
            )
        )
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
        self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_bounds_with_params(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.y = pyo.Var()
        m.p = pyo.Param(mutable=True)
        m.y.setlb(m.p)
        m.p.value = 1
        m.obj = pyo.Objective(expr=m.y)
        res = opt.solve(m)
        self.assertAlmostEqual(m.y.value, 1)
        m.p.value = -1
        res = opt.solve(m)
        self.assertAlmostEqual(m.y.value, -1)
        m.y.setlb(None)
        m.y.setub(m.p)
        m.obj.sense = pyo.maximize
        m.p.value = 5
        res = opt.solve(m)
        self.assertAlmostEqual(m.y.value, 5)
        m.p.value = 4
        res = opt.solve(m)
        self.assertAlmostEqual(m.y.value, 4)
        m.y.setub(None)
        m.y.setlb(m.p)
        m.obj.sense = pyo.minimize
        m.p.value = 3
        res = opt.solve(m)
        self.assertAlmostEqual(m.y.value, 3)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_bounds_with_immutable_params(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        # this test is for issue #2574
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.p = pyo.Param(mutable=False, initialize=1)
        m.q = pyo.Param([1, 2], mutable=False, initialize=10)
        m.y = pyo.Var()
        m.y.setlb(m.p)
        m.y.setub(m.q[1])
        m.obj = pyo.Objective(expr=m.y)
        res = opt.solve(m)
        self.assertAlmostEqual(m.y.value, 1)
        m.y.setlb(m.q[2])
        res = opt.solve(m)
        self.assertAlmostEqual(m.y.value, 10)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_solution_loader(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(1, None))
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=(0, m.y - m.x, None))
        m.c2 = pyo.Constraint(expr=(0, m.y - m.x + 1, None))
        opt.config.load_solution = False
        res = opt.solve(m)
        self.assertIsNone(m.x.value)
        self.assertIsNone(m.y.value)
        res.solution_loader.load_vars()
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 1)
        m.x.value = None
        m.y.value = None
        res.solution_loader.load_vars([m.y])
        self.assertIsNone(m.x.value)
        self.assertAlmostEqual(m.y.value, 1)
        primals = res.solution_loader.get_primals()
        self.assertIn(m.x, primals)
        self.assertIn(m.y, primals)
        self.assertAlmostEqual(primals[m.x], 1)
        self.assertAlmostEqual(primals[m.y], 1)
        primals = res.solution_loader.get_primals([m.y])
        self.assertNotIn(m.x, primals)
        self.assertIn(m.y, primals)
        self.assertAlmostEqual(primals[m.y], 1)
        if opt_class != MAiNGO:
            reduced_costs = res.solution_loader.get_reduced_costs()
            self.assertIn(m.x, reduced_costs)
            self.assertIn(m.y, reduced_costs)
            self.assertAlmostEqual(reduced_costs[m.x], 1)
            self.assertAlmostEqual(reduced_costs[m.y], 0)
            reduced_costs = res.solution_loader.get_reduced_costs([m.y])
            self.assertNotIn(m.x, reduced_costs)
            self.assertIn(m.y, reduced_costs)
            self.assertAlmostEqual(reduced_costs[m.y], 0)
            duals = res.solution_loader.get_duals()
            self.assertIn(m.c1, duals)
            self.assertIn(m.c2, duals)
            self.assertAlmostEqual(duals[m.c1], 1)
            self.assertAlmostEqual(duals[m.c2], 0)
            duals = res.solution_loader.get_duals([m.c1])
            self.assertNotIn(m.c2, duals)
            self.assertIn(m.c1, duals)
            self.assertAlmostEqual(duals[m.c1], 1)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_time_limit(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available() or opt_class == MAiNGO:
            raise unittest.SkipTest
        from sys import platform

        if platform == 'win32':
            raise unittest.SkipTest

        N = 30
        m = pyo.ConcreteModel()
        m.jobs = pyo.Set(initialize=list(range(N)))
        m.tasks = pyo.Set(initialize=list(range(N)))
        m.x = pyo.Var(m.jobs, m.tasks, bounds=(0, 1))

        random.seed(0)
        coefs = list()
        lin_vars = list()
        for j in m.jobs:
            for t in m.tasks:
                coefs.append(random.uniform(0, 10))
                lin_vars.append(m.x[j, t])
        obj_expr = LinearExpression(
            linear_coefs=coefs, linear_vars=lin_vars, constant=0
        )
        m.obj = pyo.Objective(expr=obj_expr, sense=pyo.maximize)

        m.c1 = pyo.Constraint(m.jobs)
        m.c2 = pyo.Constraint(m.tasks)
        for j in m.jobs:
            expr = LinearExpression(
                linear_coefs=[1] * N,
                linear_vars=[m.x[j, t] for t in m.tasks],
                constant=0,
            )
            m.c1[j] = expr == 1
        for t in m.tasks:
            expr = LinearExpression(
                linear_coefs=[1] * N,
                linear_vars=[m.x[j, t] for j in m.jobs],
                constant=0,
            )
            m.c2[t] = expr == 1
        if type(opt) is Ipopt:
            opt.config.time_limit = 1e-6
        else:
            opt.config.time_limit = 0
        opt.config.load_solution = False
        res = opt.solve(m)
        if type(opt) is Cbc:  # I can't figure out why CBC is reporting max iter...
            self.assertIn(
                res.termination_condition,
                {TerminationCondition.maxIterations, TerminationCondition.maxTimeLimit},
            )
        else:
            self.assertEqual(
                res.termination_condition, TerminationCondition.maxTimeLimit
            )

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_objective_changes(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.c1 = pyo.Constraint(expr=m.y >= m.x + 1)
        m.c2 = pyo.Constraint(expr=m.y >= -m.x + 1)
        m.obj = pyo.Objective(expr=m.y)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        m.obj = pyo.Objective(expr=2 * m.y)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 2)
        m.obj.expr = 3 * m.y
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 3)
        m.obj.sense = pyo.maximize
        opt.config.load_solution = False
        res = opt.solve(m)
        if opt_class != MAiNGO:
            self.assertIn(
                res.termination_condition,
                {
                    TerminationCondition.unbounded,
                    TerminationCondition.infeasibleOrUnbounded,
                },
            )
        m.obj.sense = pyo.minimize
        opt.config.load_solution = True
        m.obj = pyo.Objective(expr=m.x * m.y)
        m.x.fix(2)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 6, 6)
        m.x.fix(3)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 12, 6)
        m.x.unfix()
        m.y.fix(2)
        m.x.setlb(-3)
        m.x.setub(5)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, -2, 6)
        m.y.unfix()
        m.x.setlb(None)
        m.x.setub(None)
        m.e = pyo.Expression(expr=2)
        m.obj = pyo.Objective(expr=m.e * m.y)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 2)
        m.e.expr = 3
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 3)
        opt.update_config.check_for_new_objective = False
        m.e.expr = 4
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 4)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_domain(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(1, None), domain=pyo.NonNegativeReals)
        m.obj = pyo.Objective(expr=m.x)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        m.x.setlb(-1)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 0)
        m.x.setlb(1)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        m.x.setlb(-1)
        m.x.domain = pyo.Reals
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, -1)
        m.x.domain = pyo.NonNegativeReals
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 0)

    @parameterized.expand(input=_load_tests(mip_solvers, only_child_vars_options))
    def test_domain_with_integers(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-1, None), domain=pyo.NonNegativeIntegers)
        m.obj = pyo.Objective(expr=m.x)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 0)
        m.x.setlb(0.5)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        m.x.setlb(-5.5)
        m.x.domain = pyo.Integers
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, -5)
        m.x.domain = pyo.Binary
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 0)
        m.x.setlb(0.5)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_fixed_binaries(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var(domain=pyo.Binary)
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.y)
        m.c = pyo.Constraint(expr=m.y >= m.x)
        m.x.fix(0)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 0, 5)
        m.x.fix(1)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1, 5)

        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        opt.update_config.treat_fixed_vars_as_params = False
        m.x.fix(0)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 0, 5)
        m.x.fix(1)
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1, 5)

    @parameterized.expand(input=_load_tests(mip_solvers, only_child_vars_options))
    def test_with_gdp(
        self, name: str, opt_class: Type[PersistentSolver], only_child_vars
    ):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest

        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-10, 10))
        m.y = pyo.Var(bounds=(-10, 10))
        m.obj = pyo.Objective(expr=m.y)
        m.d1 = gdp.Disjunct()
        m.d1.c1 = pyo.Constraint(expr=m.y >= m.x + 2)
        m.d1.c2 = pyo.Constraint(expr=m.y >= -m.x + 2)
        m.d2 = gdp.Disjunct()
        m.d2.c1 = pyo.Constraint(expr=m.y >= m.x + 1)
        m.d2.c2 = pyo.Constraint(expr=m.y >= -m.x + 1)
        m.disjunction = gdp.Disjunction(expr=[m.d2, m.d1])
        pyo.TransformationFactory("gdp.bigm").apply_to(m)

        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1, 6)
        self.assertAlmostEqual(m.x.value, 0, 6)
        self.assertAlmostEqual(m.y.value, 1, 6)

        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        opt.use_extensions = True
        res = opt.solve(m)
        self.assertAlmostEqual(res.best_feasible_objective, 1, 6)
        self.assertAlmostEqual(m.x.value, 0, 6)
        self.assertAlmostEqual(m.y.value, 1, 6)

    @parameterized.expand(input=all_solvers)
    def test_variables_elsewhere(self, name: str, opt_class: Type[PersistentSolver]):
        opt: PersistentSolver = opt_class(only_child_vars=False)
        if not opt.available():
            raise unittest.SkipTest

        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.b = pyo.Block()
        m.b.obj = pyo.Objective(expr=m.y)
        m.b.c1 = pyo.Constraint(expr=m.y >= m.x + 2)
        m.b.c2 = pyo.Constraint(expr=m.y >= -m.x)

        res = opt.solve(m.b)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        self.assertAlmostEqual(m.x.value, -1)
        self.assertAlmostEqual(m.y.value, 1)

        m.x.setlb(0)
        res = opt.solve(m.b)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 2)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 2)

    @parameterized.expand(input=all_solvers)
    def test_variables_elsewhere2(self, name: str, opt_class: Type[PersistentSolver]):
        opt: PersistentSolver = opt_class(only_child_vars=False)
        if not opt.available():
            raise unittest.SkipTest

        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()

        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=m.y >= m.x)
        m.c2 = pyo.Constraint(expr=m.y >= -m.x)
        m.c3 = pyo.Constraint(expr=m.y >= m.z + 1)
        m.c4 = pyo.Constraint(expr=m.y >= -m.z + 1)

        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 1)
        sol = res.solution_loader.get_primals()
        self.assertIn(m.x, sol)
        self.assertIn(m.y, sol)
        self.assertIn(m.z, sol)

        del m.c3
        del m.c4
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 0)
        sol = res.solution_loader.get_primals()
        self.assertIn(m.x, sol)
        self.assertIn(m.y, sol)
        self.assertNotIn(m.z, sol)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_bug_1(self, name: str, opt_class: Type[PersistentSolver], only_child_vars):
        opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
        if not opt.available():
            raise unittest.SkipTest

        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(3, 7))
        m.y = pyo.Var(bounds=(-10, 10))
        m.p = pyo.Param(mutable=True, initialize=0)

        m.obj = pyo.Objective(expr=m.y)
        m.c = pyo.Constraint(expr=m.y >= m.p * m.x)

        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 0)

        m.p.value = 1
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(res.best_feasible_objective, 3)

    @parameterized.expand(input=_load_tests(all_solvers, only_child_vars_options))
    def test_bug_2(self, name: str, opt_class: Type[PersistentSolver], only_child_vars):
        """
        This test is for a bug where an objective containing a fixed variable does
        not get updated properly when the variable is unfixed.
        """
        for fixed_var_option in [True, False]:
            opt: PersistentSolver = opt_class(only_child_vars=only_child_vars)
            if not opt.available():
                raise unittest.SkipTest
            opt.update_config.treat_fixed_vars_as_params = fixed_var_option

            m = pyo.ConcreteModel()
            m.x = pyo.Var(bounds=(-10, 10))
            m.y = pyo.Var()
            m.obj = pyo.Objective(expr=3 * m.y - m.x)
            m.c = pyo.Constraint(expr=m.y >= m.x)

            m.x.fix(1)
            res = opt.solve(m)
            self.assertAlmostEqual(res.best_feasible_objective, 2, 5)

            m.x.unfix()
            m.x.setlb(-9)
            m.x.setub(9)
            res = opt.solve(m)
            self.assertAlmostEqual(res.best_feasible_objective, -18, 5)


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestLegacySolverInterface(unittest.TestCase):
    @parameterized.expand(input=all_solvers)
    def test_param_updates(self, name: str, opt_class: Type[PersistentSolver]):
        opt = pyo.SolverFactory('appsi_' + name)
        if not opt.available(exception_flag=False):
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.a1 = pyo.Param(mutable=True)
        m.a2 = pyo.Param(mutable=True)
        m.b1 = pyo.Param(mutable=True)
        m.b2 = pyo.Param(mutable=True)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.Constraint(expr=(0, m.y - m.a1 * m.x - m.b1, None))
        m.c2 = pyo.Constraint(expr=(None, -m.y + m.a2 * m.x + m.b2, 0))
        if opt_class != MAiNGO:
            m.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

        params_to_test = [(1, -1, 2, 1), (1, -2, 2, 1), (1, -1, 3, 1)]
        for a1, a2, b1, b2 in params_to_test:
            m.a1.value = a1
            m.a2.value = a2
            m.b1.value = b1
            m.b2.value = b2
            res = opt.solve(m)
            pyo.assert_optimal_termination(res)
            self.assertAlmostEqual(m.x.value, (b2 - b1) / (a1 - a2))
            self.assertAlmostEqual(m.y.value, a1 * (b2 - b1) / (a1 - a2) + b1)
            if opt_class != MAiNGO:
                self.assertAlmostEqual(m.dual[m.c1], (1 + a1 / (a2 - a1)))
                self.assertAlmostEqual(m.dual[m.c2], a1 / (a2 - a1))

    @parameterized.expand(input=all_solvers)
    def test_load_solutions(self, name: str, opt_class: Type[PersistentSolver]):
        opt = pyo.SolverFactory('appsi_' + name)
        if not opt.available(exception_flag=False):
            raise unittest.SkipTest
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.obj = pyo.Objective(expr=m.x)
        m.c = pyo.Constraint(expr=(-1, m.x, 1))
        if opt_class != MAiNGO:
            m.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
        res = opt.solve(m, load_solutions=False)
        pyo.assert_optimal_termination(res)
        self.assertIsNone(m.x.value)
        if opt_class != MAiNGO:
            self.assertNotIn(m.c, m.dual)
        m.solutions.load_from(res)
        self.assertAlmostEqual(m.x.value, -1)
        if opt_class != MAiNGO:
            self.assertAlmostEqual(m.dual[m.c], 1)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .gurobi import Gurobi, GurobiResults
from .ipopt import Ipopt
from .cbc import Cbc
from .cplex import Cplex
from .highs import Highs
from .wntr import Wntr, WntrResults
from .maingo import MAiNGO



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import enum


class ActionStatus(str, enum.Enum):
    done = 'done'
    error = 'error'
    queued = 'queued'
    executing = 'executing'
    unknown = 'unknown'

    # Overloading __str__ is needed to match the behavior of the old
    # pyutilib.enum class (removed June 2020). There are spots in the
    # code base that expect the string representation for items in the
    # enum to not include the class name. New uses of enum shouldn't
    # need to do this.
    def __str__(self):
        return self.value


def solve_all_instances(solver_manager, solver, instances, **kwds):
    """
    A simple utility to apply a solver to a list of problem instances.
    """
    solver_manager.solve_all(solver, instances, **kwds)


class ActionManagerError(Exception):
    """
    An exception used when an error occurs within an ActionManager.
    """

    def __init__(self, *args, **kargs):
        Exception.__init__(self, *args, **kargs)  # pragma:nocover


class ActionHandle:
    id_counter = 0

    def __init__(self, error=False, explanation=""):
        """Constructor"""
        if error:
            self.id = -1
        else:
            self.id = ActionHandle.id_counter
            ActionHandle.id_counter += 1
            self.status = ActionStatus.error
        self.explanation = explanation

    def update(self, ah):
        """Update the contents of the provided ActionHandle"""
        self.id = ah.id
        self.status = ah.status

    def __lt__(self, other):
        return self.id < other.id

    def __hash__(self):
        return self.id.__hash__()

    def __eq__(self, other):
        return (
            (self.__class__ is other.__class__)
            and (self.id.__hash__() == other.__hash__())
            and (self.id == other.id)
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return str(self.id)


FailedActionHandle = ActionHandle(error=True)


class AsynchronousActionManager:
    @staticmethod
    def _flatten(*args):
        ahs = set()
        if len(args) > 0:
            for item in args:
                if type(item) is ActionHandle:
                    ahs.add(item)
                elif type(item) in (list, tuple, dict, set):
                    for ah in item:
                        if type(ah) is not ActionHandle:  # pragma:nocover
                            raise ActionManagerError("Bad argument type %s" % str(ah))
                        ahs.add(ah)
                else:  # pragma:nocover
                    raise ActionManagerError("Bad argument type %s" % str(item))
        return ahs

    def __init__(self):
        """Constructor"""
        self.clear()

    def clear(self):
        """
        Clear manager state
        """
        self.event_handle = {}
        self.results = {}
        self.queued_action_counter = 0

    def execute(self, *args, **kwds):
        """
        Synchronously execute an action.
        """
        ah = self.queue(*args, **kwds)
        results = self.wait_for(ah)
        if results is None:
            raise ActionManagerError(
                "Problem executing an event.  No results are available."
            )
        return results

    def queue(self, *args, **kwds):
        """
        Queue an action, returning an ActionHandle object.
        """
        ah = ActionHandle()
        self.event_handle[ah.id] = ah
        ah.status = ActionStatus.queued
        self.queued_action_counter += 1
        return self._perform_queue(ah, *args, **kwds)

    def wait_all(self, *args):
        """
        Wait for all actions to complete.  The arguments to this method
        are expected to be ActionHandle objects or iterators that return
        ActionHandle objects.  If no arguments are provided, then this
        method will terminate after all queued actions are
        """
        #
        # Collect event handlers from the arguments
        #
        ahs = set()
        if len(args) == 0:
            ahs.update(
                ah
                for ah in self.event_handle.values()
                if ah.status == ActionStatus.queued
            )
        else:
            ahs = self._flatten(*args)
        #
        # Iterate until all ah's have completed
        #
        while len(ahs) > 0:
            ah = self.wait_any()
            ahs.discard(ah)

    def wait_any(self, *args):
        """
        Wait for any action (or any of the specified actions) to
        complete, and return the corresponding ActionHandle.
        """
        ah = None
        if len(args):
            ahs = self._flatten(*args)
            ah = None
            while ah not in ahs:
                ah = self._perform_wait_any()
        else:
            while ah is None:
                ah = self._perform_wait_any()
        if ah == FailedActionHandle:
            return ah
        self.queued_action_counter -= 1
        self.event_handle[ah.id].update(ah)
        return self.event_handle[ah.id]

    def wait_for(self, ah):
        """
        Wait for the specified action to complete.
        """
        tmp = self.wait_any()
        while tmp != ah:
            tmp = self.wait_any()
            if tmp == FailedActionHandle:
                raise ActionManagerError("Action %s failed: %s" % (ah, tmp.explanation))
        return self.get_results(ah)

    def num_queued(self):
        """
        Return the number of queued actions.
        """
        return self.queued_action_counter

    def get_status(self, ah):
        """
        Return the status of the ActionHandle.
        """
        return ah.status

    def get_results(self, ah):
        """
        Return solver results.  If solver results are not available,
        return None.
        """
        if ah.id in self.results:
            # extremely important - clean up the map before returning the result.
            # otherwise, it will be around forever, acting as a memory leak (because
            # the solver manager will always retain a reference).
            result = self.results[ah.id]
            del self.results[ah.id]
            return result
        return None

    def _perform_queue(self, ah, *args, **kwds):
        """
        Perform the queue operation.  This method returns the
        ActionHandle, and the ActionHandle status indicates whether
        the queue was successful.
        """
        raise ActionManagerError(
            "The _perform_queue method is not defined"
        )  # pragma:nocover

    def _perform_wait_any(self):
        """
        Perform the wait_any operation.  This method returns an
        ActionHandle with the results of waiting.  If None is returned
        then the ActionManager assumes that it can call this method
        again.  Note that an ActionHandle can be returned with a dummy
        value, to indicate an error.
        """
        raise ActionManagerError(
            "The _perform_wait_any method is not defined"
        )  # pragma:nocover
